"""
Routines related to halo mass function and other cosmological calculations.

.. warning ::
    These routines were implemented in 2012 when there were no other python libraries that could do this.
    Since then, cosmology-focussed libraries such as `hmf <https://hmf.readthedocs.io/en/latest/index.html>`_,
    `Colossus <https://bdiemer.bitbucket.io/colossus/index.html>`_ and
    `CCL <https://ccl.readthedocs.io/en/latest/>`_ have been developed.
    For precision cosmology applications, we recommend using these libraries, since the routines here
    have not been tested for precision applications (and are known to use approximations).

"""

import copy
import logging
import math
import os
import warnings

import numpy as np
import scipy
import scipy.interpolate

import pynbody

from .. import configuration, units, util
from . import cosmology

logger = logging.getLogger('pynbody.analysis.hmf')


#######################################################################
# Filters
#######################################################################

class FieldFilter:
    """Represents a filter acting on a field"""

    def M_to_R(self, M):
        """Return the mass scale (Msol h^-1) for a given length (Mpc h^-1 comoving)"""
        return (M / (self.gammaF * self.rho_bar)) ** 0.3333

    def R_to_M(self, R):
        """Return the length scale (Mpc h^-1 comoving) for a given spherical mass (Msol h^-1)"""
        return self.gammaF * self.rho_bar * R ** 3

    @staticmethod
    def Wk(kR):
        """Return the Fourier-space filter function, as a function of kR where R is the scale of the filter"""
        raise NotImplementedError("Not implemented")


class TophatFilter(FieldFilter):
    """A top-hat filter in real space"""
    def __init__(self, context):
        self.gammaF = 4 * math.pi / 3
        self.rho_bar = cosmology.rho_M(context, unit="Msol Mpc^-3 h^2 a^-3")

    @staticmethod
    def Wk(kR):
        return 3 * (np.sin(kR) - kR * np.cos(kR)) / (kR) ** 3


class GaussianFilter(FieldFilter):
    """A Gaussian filter"""

    def __init__(self, context):
        self.gammaF = (2 * math.pi) ** 1.5
        self.rho_bar = cosmology.rho_M(context, unit="Msol Mpc^-3 h^2 a^-3")

    @staticmethod
    def Wk(kR):
        return np.exp(-(kR) ** 2 / 2)


class HarmonicStepFilter(FieldFilter):
    """A step filter in harmonic space"""

    def __init__(self, context):
        self.gammaF = 6 * math.pi ** 2
        self.rho_bar = cosmology.rho_M(context, unit="Msol Mpc^-3 h^2 a^-3")

    @staticmethod
    def Wk(kR):
        return (kR < 1)

#######################################################################
# Power spectrum management / normalization
#######################################################################


class PowerSpectrum:
    """A power spectrum object, which can be called to return the power spectrum at a given wavenumber."""

    def __init__(self, context, filename=None, k=None, Pk=None, log_interpolation=True):
        """Set up a power spectrum object.

        Either a tabulated power spectrum or the filename of a tabulated power spectrum can be passed.
        If neither is provided, a default power spectrum is used which assumes Planck 2018 values.

        Parameters
        ----------

        context : SimSnap
            The simulation snapshot from which to pull the cosmological context. This is used for
            the linear growth factor and sigma8 normalization.

        filename : str, optional
            The name of a file containing the power spectrum.

        k : array, optional
            The wavenumbers of the power spectrum in comoving h/Mpc

        Pk : array, optional
            The power spectrum in Mpc^3 h^-3

        log_interpolation : bool, optional
            If True, interpolate in log space. This is generally more accurate for power spectra.

        """

        if k is None or Pk is None:
            if filename is None:
                warnings.warn(
                    "Using the default power-spectrum spectrum which assumes Planck 2018 values: "
                    "Omega_c h^2 = 0.120; Omega_b h^2 = 0.0224; h = 0.676; n_s = 0.9667. Sigma8 will be correctly scaled to the simulation value. ",
                    RuntimeWarning)
                filename = os.path.join(os.path.dirname(__file__), "CAMB_Planck18")
            k, Pk = np.loadtxt(filename, unpack=True)

        self._orig_k_min = k.min()
        self._orig_k_max = k.max()

        bot_k = 1.e-4

        if k[0] > bot_k:
            # extrapolate out
            n = math.log10(Pk[1] / Pk[0]) / math.log10(k[1] / k[0])

            Pkinterp = 10 ** (math.log10(Pk[0]) - math.log10(k[0] / bot_k) * n)
            k = np.hstack((bot_k, k))
            Pk = np.hstack((Pkinterp, Pk))

        top_k = 1.e4

        if k[-1] < top_k:
            # extrapolate out
            n = math.log10(Pk[-1] / Pk[-2]) / math.log10(k[-1] / k[-2])

            Pkinterp = 10 ** (math.log10(Pk[-1]
                                         ) - math.log10(k[-1] / top_k) * n)
            k = np.hstack((k, top_k))
            Pk = np.hstack((Pk, Pkinterp))

        self.k = k.view(pynbody.array.SimArray)
        self.k.units = "Mpc^-1 h a^-1"

        self.Pk_z0_unnormalised = Pk.view(pynbody.array.SimArray)
        self.Pk_z0_unnormalised.units = "Mpc^3 h^-3"

        self.k.sim = context
        self.Pk_z0_unnormalised.sim = context

        self._lingrowth = 1

        if context.properties['z'] != 0:
            self._lingrowth = cosmology.linear_growth_factor(context) ** 2

        self._default_filter = TophatFilter(context)
        self.min_k = self.k.min()*1.000001
        self.max_k = self.k.max()*0.999999
        self._norm = 1

        self._log_interp = log_interpolation

        self._init_interpolation()

        self.set_sigma8(context.properties['sigma8'])

    def _init_interpolation(self):
        if self._log_interp:
            self._interp = scipy.interpolate.interp1d(
                np.log(self.k), np.log(self.Pk_z0_unnormalised))
        else:
            self._interp = scipy.interpolate.interp1d(np.log(self.k), self.Pk_z0_unnormalised)

    def set_sigma8(self, sigma8):
        """Update the normalisation for a given sigma8 value at z=0"""
        current_sigma8_2 = self.get_sigma8() ** 2
        self._norm *= sigma8 ** 2 / current_sigma8_2

    def get_sigma8(self):
        """Calculate the current sigma8 value at z=0"""
        current_sigma8 = math.sqrt(
            variance(8.0, self._default_filter, self, True) / self._lingrowth)
        return current_sigma8

    def __call__(self, k):
        """Evaluate the power spectrum at the snapshot redshift"""
        if self._log_interp:
            return self._norm * self._lingrowth * np.exp(self._interp(np.log(k)))
        else:
            return self._norm * self._lingrowth * self._interp(np.log(k))


class PowerSpectrumCAMB(PowerSpectrum):
    """A power spectrum object that calculates the power spectrum on the fly using CAMB.

    This is slower than using a precomputed power spectrum, but is more flexible."""

    def __init__(self, context, use_context=True, camb_params={}, log_interpolation=True):
        """Runs CAMB to calculate a power spectrum for the provided context.

        Parameters
        ----------
        context : SimSnap
            The simulation snapshot from which to pull the cosmological context.

        use_context : bool, optional
            If True, use the context to set the cosmological parameters in CAMB.
            If False, use only the parameters in ``camb_params``.

        camb_params : dict, optional
            A dictionary of parameters to pass to CAMB's ``set_cosmology`` method. Note that the
            scalar spectral index ``ns`` is not included in ``set_cosmology``; if provided here, it is
            extracted and passed to ``InitPower.set_params``.

        log_interpolation : bool, optional
            If True, interpolate in log space. This is generally more accurate for power spectra.

        """

        try:
            import camb
        except ImportError:
            warnings.warn("The camb module is not installed. Falling back to using a pre-computed power spectrum"
                          "which may not be appropriate for your cosmology.")
            super().__init__(context)
            return

        camb_params_obj = camb.CAMBparams()
        camb_params = copy.copy(camb_params)

        # default value for ns, in case use_context is false:
        ns = configuration.config_parser.get('default-cosmology', 'ns')

        if use_context:
            h0 = context.properties['h']
            omBh2 = context.properties['omegaB0'] * h0 ** 2
            omMh2 = context.properties['omegaM0'] * h0 ** 2
            if omBh2 == 0.0 and 'ombh2' not in camb_params_obj:
                warnings.warn("OmegaB0 is zero, presumably because this is a DMO run. For the power spectrum, setting ombh2 to 0.0224 (Planck 2018)."
                              "To override, set 'omegaB0' in the simulation parameters, or 'ombh2' in the camb_params argument.")
                omBh2 = 0.0224
            OmCh2 = omMh2 - omBh2

            if 'H0' not in camb_params:
                camb_params['H0'] = h0 * 100
            if 'ombh2' not in camb_params:
                camb_params['ombh2'] = omBh2
            if 'omch2' not in camb_params:
                camb_params['omch2'] = OmCh2

            ns = context.properties['ns']

        if 'ns' in camb_params:
            ns = camb_params.pop('ns')

        camb_params_obj.set_cosmology(**camb_params)

        camb_params_obj.InitPower.set_params(ns=ns)

        camb_params_obj.WantTransfer = True

        # always get the matter power spectrum at z=0; growth factor will be scaled later if needed
        camb_params_obj.set_matter_power(redshifts=[0.0], kmax=1e2, k_per_logint=0,
                                     silent=True, nonlinear=False)
        results = camb.get_results(camb_params_obj)

        k, z, Pk = results.get_matter_power_spectrum(minkh=1.e-4, maxkh=1e2, npoints=1000)

        # filename = os.path.join(os.path.dirname(__file__), "CAMB_Planck18")
        # np.savetxt(filename, np.array([k, Pk[0]]).T)

        assert len(z)==1
        assert z[0] == 0.0

        super().__init__(context, k=k, Pk=Pk[0], log_interpolation=log_interpolation)



class BiasedPowerSpectrum(PowerSpectrum):
    """A power spectrum object with linear bias"""

    def __init__(self, bias, pspec):
        """Set up a biased power spectrum.

        Parameters
        ----------
        bias : float  or function
            Either a constant linear bias factor, or a function of wavenumber that returns the bias at that wavenumber.

        pspec : PowerSpectrum
            The power spectrum object to wrap with a bias
        """

        if not hasattr(bias, '__call__'):
            bias = lambda x: bias

        self._bias = bias
        self._pspec = pspec
        self._norm = 1.0
        self.min_k = pspec.min_k
        self.max_k = pspec.max_k
        self.k = pspec.k
        self.Pk_z0_unnormalised = pspec.Pk_z0_unnormalised * self._bias(self.k) ** 2

    def __call__(self, k):
        return self._norm * self._pspec(k) * self._bias(k) ** 2


#######################################################################
# Variance calculation
#######################################################################

def variance(M_or_R, f_filter=TophatFilter, powspec=PowerSpectrum, arg_is_R=False):
    """Calculate the variance of the density field smoothed on a mass scale M, or optionally a length scale R.

    Parameters
    ----------
    M_or_R : float or array
        The mass scale in Msol/h or the radius of the filter in Mpc/h comoving. If an array, the
        variance is calculated for each element of the array.

    f_filter : FieldFilter, optional
        The filter to use. Default is a top-hat filter.

    powspec : PowerSpectrum, optional
        The power spectrum object to use. Default is a Planck 2018 power spectrum at z=0.

    arg_is_R : bool, optional
        If True, interpret the input as a length scale R rather than a mass scale M.

    Returns
    -------
    float or array
        The variance of the density field smoothed on the given scale(s).

    """
    if hasattr(M_or_R, '__len__'):
        ax = pynbody.array.SimArray(
            [variance(Mi, f_filter, powspec, arg_is_R) for Mi in M_or_R])
        # hopefully dimensionless
        ax.units = powspec.Pk_z0_unnormalised.units * powspec.k.units ** 3
        return ax

    if arg_is_R:
        R = M_or_R
    else:
        R = f_filter.M_to_R(M_or_R)

    integrand = lambda k: k ** 2 * powspec(k) * f_filter.Wk(k * R) ** 2
    integrand_ln_k = lambda k: np.exp(k) * integrand(np.exp(k))
    v =  scipy.integrate.quad(integrand_ln_k, math.log(powspec.min_k), math.log(
        1. / R) + 3, epsrel=1.e-6)[0] / (2 * math.pi ** 2)

    return v


def get_neffm(mass, sigma):
    """Calculate the effective spectral index of the power spectrum at a given mass scale."""
    dlnm = np.diff(np.log(mass))
    dlnsigmainv = np.diff(np.log(1. / sigma))
    neff = 6. * dlnsigmainv / dlnm - 3.
    return neff


@units.takes_arg_in_units((0, "Mpc h^-1"))
def correlation(r, powspec=PowerSpectrum):
    """Calculate the correlation function of the density field at a specified radius."""

    if hasattr(r, '__len__'):
        ax = pynbody.array.SimArray([correlation(ri,  powspec) for ri in r])
        ax.units = powspec.Pk_z0_unnormalised.units * powspec.k.units ** 3
        return ax

    # Because sin kr becomes so highly oscilliatory, normal
    # quadrature is slow/inaccurate for this problem. The following
    # is the best way I could come up with to overcome that.
    #
    # For small kr, sin kr/kr is represented as a Taylor expansion and
    # each segment of the power spectrum is integrated over, summing
    # over the Taylor series to convergence.
    #
    # When the convergence of this starts to fail, each segment of the
    # power spectrum is still represented by a power law, but the
    # exact integral boils down to a normal incomplete gamma function
    # extended into the complex plane.
    #
    # Originally, we had:
    #
    # integrand = lambda k: k**2 * powspec(k) * (np.sin(k*r)/(k*r))
    # integrand_ln_k = lambda k: np.exp(k)*integrand(np.exp(k))
    #

    tot = 0
    defer = False

    k = powspec.k

    gamma_method = False

    for k_bot, k_top in zip(k[:-1], k[1:]):

        if k_bot >= k_top:
            continue

        # express segment as P(k) = P0*k^n
        Pk_top = powspec(k_top)
        Pk_bot = powspec(k_bot)

        n = np.log(Pk_top / Pk_bot) / np.log(k_top / k_bot)
        P0 = Pk_top / k_top ** n

        if n != n or abs(n) > 2:
            # looks nasty in log space, so interpolate linearly instead
            grad = (Pk_top - Pk_bot) / (k_top - k_bot)
            segment = ((-2 * grad + k_bot * Pk_bot * r ** 2) * np.cos(k_bot * r) +
                       (2 * grad - k_top * (-(grad * k_bot) + grad * k_top + Pk_bot) * r ** 2) * np.cos(k_top * r) -
                       (grad * k_bot + Pk_bot) * r * np.sin(k_bot * r) +
                       (-(grad * k_bot) + 2 * grad * k_top + Pk_bot) * r * np.sin(k_top * r)) / r ** 4

        elif k_top * r < 6.0 and not gamma_method:
            # approximate sin y/y as polynomial = \sum_m coeff_m y^m

            segment = 0
            term = 0

            m = 0
            coeff = 1
            while m == 0 or (abs(term / segment) > 1.e-7 and m < 50):
                if m > 0:
                    coeff *= (-1.0) / (m * (m + 1))

                # integral is P0 * r^m * int_(k_bot)^(k_top) k^(2+n+m) dk = P0
                # r^m [k^(3+n+m)/(3+n+m)]
                top_val = k_top ** (3 + n + m) / (3 + n + m)
                bot_val = k_bot ** (3 + n + m) / (3 + n + m)
                term = P0 * (r ** m) * (top_val - bot_val) * coeff
                segment += term
                m += 2

            if m >= 50:
                raise RuntimeError("Convergence failure in sin y/y series integral")

            if m > 18:
                gamma_method = True
                # experience suggests when you have to sum beyond m=18, it's faster
                # to switch to the method below

        else:

            # now integral of this segment is exactly
            # P0 * int_(k_bot)^(k_top) k^(2+n) sin(kr)/(kr) = (P0/r^(n+3)) Im[ (i)^(-n-2) Gamma(n+2,i k_bot r, i k_top r)]
            # First we need to evaluate the Gamma integral sufficiently
            # accurately

            top_val = util.gamma_inc(n + 2, (1.0j) * r * k_top)
            bot_val = util.gamma_inc(n + 2, (1.0j) * r * k_bot)
            segment = - \
                ((1.0j) ** (-n - 2) * P0 *
                 (top_val - bot_val) / r ** (n + 3)).imag

        tot += segment

    tot /= (2 * math.pi ** 2)

    return tot


def correlation_func(context, log_r_min=-3, log_r_max=2, delta_log_r=0.2,
                     pspec=PowerSpectrumCAMB):
    """
    Calculate the linear density field correlation function.

    Parameters
    ----------

    context : SimSnap
        The snapshot from which to pull the cosmological context

    log_r_min : float, optional
        log10 of the minimum separation (Mpc h^-1) to consider

    log_r_max : float, optional
        log10 of the maximum separation (Mpc h^-1) to consider

    delta_log_r : float, optional
        The value spacing in dex

    pspec : PowerSpectrum, optional
        The power spectrum class to use, or if an instance is provided, the power spectrum to use.
        The default is to use PowerSpectrumCAMB which will attempt to calculate a power spectrum
        from the simulation context; if it fails, it will fall back to using Planck2018 power spectrum
        and issue a warning.

    Returns
    -------

    r : array
        The separation values in Mpc h^-1

    Xi : array
        The correlation function at each separation


    """

    if isinstance(pspec, type):
        pspec = pspec(context)

    r = (10.0 ** np.arange(log_r_min, log_r_max + delta_log_r / 2,
                           delta_log_r)).view(pynbody.array.SimArray)
    r.sim = context
    r.units = "Mpc h^-1 a"

    Xi_r = np.array([correlation(ri, pspec)
                     for ri in r]).view(pynbody.array.SimArray)
    Xi_r.sim = context
    Xi_r.units = ""

    return r, Xi_r

#######################################################################
# Default kernels for halo mass function
#######################################################################


def _f_press_schechter(nu):
    """

    The Press-Schechter kernel used by halo_mass_function

    """

    f = math.sqrt(2. / math.pi) * nu * np.exp(-nu * nu / 2.)
    return f


def _f_sheth_tormen(nu, Anorm=0.3222, a=0.707, p=0.3):
    """
    Sheth & Tormen (1999) fit (see also Sheth Mo & Tormen 2001)
    """
    #  Anorm: normalization, set so all mass is in halos (integral [f nu dn]=1)
    #  a: affects mainly the number of massive halo,
    #  a=0.75 is favored by Sheth & Tormen (2002)

    f = Anorm * math.sqrt(2. * a / math.pi) * \
        (1. + np.power((1. / a / nu / nu), p))
    f *= nu * np.exp(-a * nu * nu / 2.)
    return f


def _f_jenkins(nu, deltac=1.68647):
    # Jenkins et al (2001) fit   ##  valid for  -1.2 << ln(1/sigma) << 1.05
    sigma = deltac / nu
    lnsigmainv = np.log(1. / sigma)
    if ((np.any(lnsigmainv < -1.2)) or (np.any(lnsigmainv > 1.05))):
        logger.warning(
            "Jenkins mass function is outsie of valid mass range.  Continuing calculations anyway.")
    f = 0.315 * np.exp(-np.power((np.fabs(lnsigmainv + 0.61)), 3.8))
    return f


def _f_warren(nu, deltac=1.68647):
    #  Warren et al. 2006  -- valid for (10**10 - 10**15 Msun/h)
    sigma = deltac / nu
    A = 0.7234
    a = 1.625
    b = 0.2538
    c = 1.1982
    f = A * (np.power(sigma, -a) + b) * np.exp(-c / sigma ** 2)
    return f


def _f_reed_no_z(nu, deltac=1.68647):  # universal form
    # Reed et al. (2007) fit, eqn. 9 -- with no redshift depedence (simple
    # universal form)
    """ modified S-T fit  by the G1 gaussian term and c"""
    sigma = deltac / nu
    # normalization that all mass is in halos not strictly conserved here
    Anorm = 0.3222
    a = 0.707  # affects mostly the number of massive halos
    # a=0.75    #  favored by Sheth & Tormen (2002)
    p = 0.3
    c = 1.08
    nu = deltac / sigma
    lnsigmainv = np.log(1. / sigma)
    G1 = np.exp(-np.power((lnsigmainv - 0.4), 2) / (2. * 0.6 * 0.6))
    f = Anorm * np.sqrt(2. * a / np.pi) * \
        (1. + np.power((1. / a / nu / nu), p) + 0.2 * G1)
    f *= nu * np.exp(-c * a * nu * nu / 2.)
    return f


def _f_reed_z_evo(nu, neff, deltac=1.68647):  # non-universal form
    # Reed et al. (2007) fit, eqn. 11 -- with redshift depedence for accuracy
    # at z >~ z_reion
    """ modified S-T fit  by the n_eff dependence and the G1 and G2 gaussian terms and c
    where   P(k) proportional to k_halo**(n_eff)  and
    k_halo = Mhalo / r_halo_precollapse.
    eqn 13 of Reed et al 2007   estimtes neff = 6 d ln(1/sigma(M))/ d ln M  - 3 """
    sigma = deltac / nu
    # normalization that all mass is in halos not strictly conserved here
    Anorm = 0.3222
    a = 0.707  # affects mostly the number of massive halos
    # a=0.75    #  favored by Sheth & Tormen (2002)
    p = 0.3
    c = 1.08
    nu = deltac / sigma
    lnsigmainv = np.log(1. / sigma)
    G1 = np.exp(- np.power((lnsigmainv - 0.4), 2) / (2. * 0.6 * 0.6))
    G2 = np.exp(- np.power((lnsigmainv - 0.75), 2) / (2. * 0.2 * 0.2))
    f = Anorm * np.sqrt(2. * a / np.pi) * (1. +
                                           np.power((1. / a / nu / nu), p) + 0.6 * G1 + 0.4 * G2)
    f *= nu * \
        np.exp(-c * a * nu * nu / 2. - 0.03 / (neff + 3)
               ** 2 * np.power(nu, 0.6))
    return f


def _f_bhattacharya(nu, red, deltac=1.68647):
    # Bhattacharya et al. 2010  -- 6x10**11 - 310**15 Msun/h  z=0-2
    sigma = deltac / nu
    A = 0.333 / pow((1. + red), 0.11)
    a = 0.788 / pow((1. + red), 0.01)
    p = 0.807 / pow((1. + red), 0.0)
    q = 1.795 / pow((1. + red), 0.0)
    f = A * np.sqrt(2. / np.pi) * (1. + np.power((1. / a / nu / nu), p))
    f *= np.power(nu * math.sqrt(a), q) * np.exp(-a * nu * nu / 2.)
    return f

def _cole_kaiser_bias(nu, delta_c):
    """

    The Cole-Kaiser (1989) bias function. Also in Mo & White 1996.

    """
    return 1 + (nu ** 2 - 1) / delta_c


def _sheth_tormen_bias(nu, delta_c,
                       a=0.707, b=0.5, c=0.6):
    """

    The Sheth-Tormen (1999) bias function [eq 8]

    """

    root_a = math.sqrt(a)

    return 1. + (root_a * a * nu ** 2 + root_a * b * (a * nu ** 2) ** (1. - c)
                 - (a * nu ** 2) ** c / ((a * nu ** 2) ** c + b * (1 - c) * (1 - c / 2))) \
        / (root_a * delta_c)



def halo_mass_function(context,
                       log_M_min=8.0, log_M_max=15.0, delta_log_M=0.1,
                       kern="ST",
                       pspec=PowerSpectrum,
                       delta_crit=1.686):
    """
    Returns the halo mass function, dN/d log_{10} M in units of Mpc^-3 h^3.

    See :ref:`hmf_tutorial` for an example and more information.


    Parameters
    ----------

    context : SimSnap
        The snapshot from which to pull the cosmological context. Sigma8 normalization and growth function
        redshift dependence is always taken into account. If `camb <https://camb.readthedocs.io/en/latest/>`_ is
        installed, an appropriate power spectrum is calculated for other cosmological parameters (e.g. baryon
        density, Hubble constant) from the simulation. Otherwise, a default Planck18 power spectrum is
        used.

    log_M_min : float, optional
        The minimum halo mass (Msol h^-1) to consider. Default is 8.

    log_M_max : float, optional
        The maximum halo mass (Msol h^-1) to consider. Default is 15.

    delta_log_M : float, optional
        The bin spacing of halo masses. Default is 0.1.

    kern : str or function, optional
        The kernel function which dictates what type of mass function to calculate. Default is "ST" (Sheth-Tormen).
        Other options are "PS" (Press-Schechter), "J" (Jenkins), "W" (Warren), "REEDZ" (Reed et al 2007 with redshift
        dependence), "REEDU" (Reed et al 2007 without redshift dependence), "B" (Bhattacharya).

    pspec : PowerSpectrumCAMB, optional
        A power spectrum object (which also defines the window function), overriding the cosmological
        parameters of the simulation.

    delta_crit : float, optional
        The critical overdensity for collapse. Default is 1.686.

    Returns
    -------

    M : SimArray
        The centre of the mass bins, in Msol h^-1.

    sigma : SimArray
        The linear variance of the corresponding sphere.

    N : SimArray
        The abundance of halos of that mass (Mpc^-3 h^3 comoving, per decade of mass).

    Notes
    -----

    Because numerical derivatives are involved, the value of ``delta_log_M`` affects the accuracy.

    The halo mass function code in pynbody was implemented in 2012 when there
    were no other python libraries that could do this.
    Since then, cosmology-focussed libraries such as `hmf <https://hmf.readthedocs.io/en/latest/index.html>`_,
    and `CCL <https://halotools.readthedocs.io/en/latest/>`_ have been developed.
    For precision cosmology applications, we recommend using these libraries.

    The functionality here is retained for quick cross-checks of simulations.

    """

    if isinstance(kern, str):
        kern = {'PS': _f_press_schechter,
                'ST': _f_sheth_tormen,
                'J': _f_jenkins,
                'W': _f_warren,
                'REEDZ': _f_reed_z_evo,
                # Reed et al 2007 without redshift dependence
                'REEDU': _f_reed_no_z,
                'B': _f_bhattacharya}[kern]

    rho_bar = cosmology.rho_M(context, unit="Msol Mpc^-3 h^2")
    red = context.properties['z']  # redshift is always set by simulation

    M = np.arange(log_M_min, log_M_max, delta_log_M)
    M_mid = np.arange(
        log_M_min + delta_log_M / 2, log_M_max - delta_log_M / 2, delta_log_M)

    if isinstance(pspec, type):
        pspec = pspec(context)

    sig = variance(10 ** M, pspec._default_filter, pspec)  # sigma(m)**2

    nu = delta_crit / np.sqrt(sig)
    nu.units = "1"

    nu_mid = (nu[1:] + nu[:-1]) / 2

    d_ln_nu_d_ln_M = np.diff(np.log10(nu)) / delta_log_M

    dM = np.diff(10 ** M)

    if (kern == _f_reed_z_evo):
        neff = get_neffm(10. ** M, sig ** 0.5)
        out = (rho_bar / (10 ** M_mid)) * kern(nu_mid, neff) * \
            d_ln_nu_d_ln_M * math.log(10.) * context.properties['a'] ** 3
    elif (kern == _f_bhattacharya):
       # eq 7.46, Mo, van den Bosch and White
        out = (rho_bar / (10 ** M_mid)) * kern(nu_mid, red) * \
            d_ln_nu_d_ln_M * math.log(10.) * context.properties['a'] ** 3
    else:
        # eq 7.46, Mo, van den Bosch and White
        out = (rho_bar / (10 ** M_mid)) * kern(nu_mid) * \
            d_ln_nu_d_ln_M * math.log(10.) * context.properties['a'] ** 3
    out.units = "Mpc^-3 h^3 a^-3"
    out.sim = context

    M_mid = (10 ** M_mid).view(pynbody.array.SimArray)
    M_mid.units = "Msol h^-1"
    M_mid.sim = context

    # interpolate sigma for output checking purposes
    sig = (sig[1:] + sig[:-1]) / 2

    return M_mid, np.sqrt(sig), out


@units.takes_arg_in_units((1, "Msol h^-1"), context_arg=0)
def halo_bias(context, M, kern="CK", pspec=PowerSpectrumCAMB,
              delta_crit=1.686):
    """
    Return the halo bias for the given halo mass.

    Parameters
    ----------

    context : SimSnap
        The snapshot from which to pull the cosmological context

    M : float
        The halo mass in Msol h^-1

    kern : str or function, optional
        The kernel function describing the halo bias. Default is "CK" (Cole-Kaiser).
        Other options are "ST" (Sheth-Tormen). Alternatively a function can be provided.

    pspec : PowerSpectrum, optional
        The power spectrum class to use, or if an instance is provided, the power spectrum to use.
        Default is PowerSpectrumCAMB which will attempt to calculate a power spectrum from the simulation context;
        if it fails, it will fall back to using Planck2018 power spectrum and issue a warning.

    delta_crit : float, optional
        The critical overdensity for collapse. Default is 1.686.

    Returns
    -------

    float
        The halo bias for the given halo mass.

    """

    if isinstance(kern, str):
        kern = {'CK': _cole_kaiser_bias,
                'ST': _sheth_tormen_bias}[kern]

    if isinstance(pspec, type):
        pspec = pspec(context)

    sig = variance(M, pspec._default_filter, pspec)
    nu = delta_crit / np.sqrt(sig)

    return kern(nu, delta_crit)


def simulation_halo_mass_function(snapshot_or_cat,
                                  log_M_min=8.0, log_M_max=15.0, delta_log_M=0.1,
                                  masses=None, mass_property=None, calculate_err=True,
                                  subsample_catalogue=None):
    """Estimate a halo mass function from a simulation, via binning haloes in mass.

    Parameters
    ----------

    snapshot_or_cat : SimSnap or HaloCatalogue
        The snapshot from which to calculate the halo mass function, or the halo catalogue itself.
        If a snapshot is provided, ``.halos()`` is called to get the first available catalogue.

    log_M_min : float, optional
        The minimum halo mass (Msol h^-1) to consider. Default is 8.

    log_M_max : float, optional
        The maximum halo mass (Msol h^-1) to consider. Default is 15.

    delta_log_M : float, optional
        The bin spacing of halo masses. Default is 0.1.

    masses : array, optional
        Provide an array of halo masses in Msol h^-1. If None, this is calculated from the snapshot.

    mass_property : str, optional
        The property name giving the mass of each halo. If None, sums the mass in particles in each halo
        according to the particle data in the catalogue. Summing masses can be slow for large catalogues.

    calculate_err : bool, optional
        If True, estimates error bars according to Poisson statistics.

    subsample_catalogue : int, optional
        Subsample the halo catalogue by the specified factor. Not recommended except for debugging.

    Returns
    -------

    bin_centers : SimArray
        The centre of the mass bins, in Msol h^-1.

    num_den : SimArray
        The binned number density of haloes in this snapshot in comoving Mpc**-3 h**3 per decade of mass.

    err : SimArray
        The error on the number density, if ``calculate_err`` is True.

    """

    if isinstance(snapshot_or_cat, pynbody.snapshot.SimSnap):
        snapshot = snapshot_or_cat
        halo_catalogue = snapshot.halos()
    elif isinstance(snapshot_or_cat, pynbody.halo.HaloCatalogue):
        halo_catalogue = snapshot_or_cat
        snapshot = halo_catalogue.base
    else:
        raise TypeError("snapshot_or_cat must be a SimSnap or HaloCatalogue")


    #Check that the mass resolution is uniform, this method does not handle otherwise
    if len(set(snapshot.d['mass'])) > 1:
        warnings.warn( "The mass resolution of the snapshot is not uniform (e.g. zooms). This method"
                       "will not generate a correct HMF in this case.")

    nbins = int(1 + (log_M_max - log_M_min)/delta_log_M)
    bins = np.logspace(log_M_min, log_M_max, num=nbins, base=10)
    bin_centers = (bins[:-1] + bins[1:]) / 2

    halo_catalogue.load_all()

    if subsample_catalogue is not None:
        halo_catalogue_underlying = halo_catalogue # keep alive for the lifetime of the function
        halo_catalogue = halo_catalogue[::subsample_catalogue]

    if masses is None:
        if mass_property is None:
            warnings.warn("Halo finder masses not provided. Calculating them (might take a while...)")
            masses = np.array([h['mass'].sum().in_units('1 h**-1 Msol') for h in halo_catalogue])
        else:
            masses = halo_catalogue.get_properties_all_halos(with_units=True)[mass_property]
            if units.has_unit(masses):
                masses = masses.in_units('1 h**-1 Msol')

    if np.amax(masses) > 10**log_M_max or np.amin(masses) < 10**log_M_min :
        warnings.warn("Your bin range does not encompass the full range of halo masses")

    # Calculate number of halos in each bin
    num_halos = np.histogram(masses, bins)[0]

    normalisation = (snapshot.properties['boxsize'].in_units(' a h**-1 Mpc', **snapshot.conversion_context()) ** 3)\
                    * delta_log_M

    if calculate_err:
        # Calculate error bars assuming Poisson distribution in each bin
        err = np.sqrt(num_halos )/normalisation

    num_halos = num_halos  / normalisation


    # Make sure units are consistent
    bin_centers = bin_centers.view(pynbody.array.SimArray)
    bin_centers.units = "Msol h**-1"
    bin_centers.sim = snapshot

    num_halos = num_halos.view(pynbody.array.SimArray)
    num_halos.units = "a**-3 Mpc**-3 h**3"
    num_halos.sim = snapshot

    if calculate_err:
        err = err.view(pynbody.array.SimArray)
        err.units = "a**-3 Mpc**-3 h**3"
        err.sim = snapshot
        return bin_centers, num_halos, err
    else:
        return bin_centers, num_halos



import matplotlib.pylab as plt

import pynbody

# load the snapshot and set to physical units
s = pynbody.load('testdata/gasoline_ahf/g15784.lr.01024.gz')
s.physical_units()

# load the halos
h = s.halos()

# center on the largest halo and align the disk
pynbody.analysis.angmom.sideon(h[1])

#create an image using the default bands (i, v, u)
pynbody.plot.stars.render(s,width='20 kpc')



cimport cython
cimport numpy as np

import sys

import numpy as np

from cython cimport floating


cdef extern from "math.h":
    int floor(double)nogil

ctypedef fused fused_input_type_1:
    np.float32_t
    np.float64_t

ctypedef fused fused_input_type_2:
    np.float32_t
    np.float64_t

ctypedef fused fused_input_type_3:
    np.float32_t
    np.float64_t


@cython.boundscheck(False)
@cython.wraparound(False)
def interpolate3d(int n,
                  np.ndarray[fused_input_type_1,ndim=1] x,
                  np.ndarray[fused_input_type_2,ndim=1] y,
                  np.ndarray[fused_input_type_3,ndim=1] z,
                  int n_x_vals, np.ndarray[np.float64_t,ndim=1] x_vals,
                  int n_y_vals, np.ndarray[np.float64_t,ndim=1] y_vals,
                  int n_z_vals, np.ndarray[np.float64_t,ndim=1] z_vals,
                  np.ndarray[np.float64_t,ndim=3] vals,
                  np.ndarray[np.float64_t,ndim=1] result_array) :

    from cython.parallel cimport prange

    cdef int x_top_ind, x_bot_ind, y_top_ind, y_bot_ind, z_top_ind, z_bot_ind, mid_ind
    cdef double x_fac, y_fac, z_fac
    cdef double v0, v1, v00, v01, v10, v11, v000, v001, v010, v011, v100, v101, v110, v111
    cdef double xi, yi, zi
    cdef Py_ssize_t i

    for i in prange(n,nogil=True) :
        if n_x_vals > 0 :
            xi = x[i]
        yi = y[i]
        zi = z[i]

        if n_x_vals > 0 :
            # find x indices
            x_top_ind = n_x_vals - 1
            x_bot_ind = 0

            while(x_top_ind > x_bot_ind + 1) :
                mid_ind = floor((x_top_ind-x_bot_ind)/2)+x_bot_ind
                if (xi > x_vals[mid_ind]) :
                    x_bot_ind = mid_ind
                else :
                    x_top_ind = mid_ind

        else :
            x_top_ind = 0
            x_bot_ind = 0

        # find y indices
        y_top_ind = n_y_vals - 1
        y_bot_ind = 0

        while(y_top_ind > y_bot_ind + 1) :
            mid_ind = floor((y_top_ind-y_bot_ind)/2)+y_bot_ind
            if (yi > y_vals[mid_ind]) :
                y_bot_ind = mid_ind
            else :
                y_top_ind = mid_ind

        # find z indices
        z_top_ind = n_z_vals - 1
        z_bot_ind = 0

        while(z_top_ind > z_bot_ind + 1) :
            mid_ind = floor((z_top_ind-z_bot_ind)/2)+z_bot_ind
            if (zi > z_vals[mid_ind]) :
                z_bot_ind = mid_ind
            else :
                z_top_ind = mid_ind

        if n_x_vals > 0 :
            x_fac = (xi - x_vals[x_bot_ind])/(x_vals[x_top_ind] - x_vals[x_bot_ind])

        y_fac = (yi - y_vals[y_bot_ind])/(y_vals[y_top_ind] - y_vals[y_bot_ind])
        z_fac = (zi - z_vals[z_bot_ind])/(z_vals[z_top_ind] - z_vals[z_bot_ind])

        # vertex values
        if n_x_vals > 0 :
            v000 = vals[x_bot_ind,y_bot_ind,z_bot_ind]
            v001 = vals[x_bot_ind,y_bot_ind,z_top_ind]
            v010 = vals[x_bot_ind,y_top_ind,z_bot_ind]
            v011 = vals[x_bot_ind,y_top_ind,z_top_ind]
            v100 = vals[x_top_ind,y_bot_ind,z_bot_ind]
            v101 = vals[x_top_ind,y_bot_ind,z_top_ind]
            v110 = vals[x_top_ind,y_top_ind,z_bot_ind]
            v111 = vals[x_top_ind,y_top_ind,z_top_ind]

            v00 = v000*(1.0-x_fac) + v100*x_fac
            v10 = v010*(1.0-x_fac) + v110*x_fac
            v01 = v001*(1.0-x_fac) + v101*x_fac
            v11 = v011*(1.0-x_fac) + v111*x_fac

        else :
            v00 = vals[0, y_bot_ind, z_bot_ind]
            v01 = vals[0, y_bot_ind, z_top_ind]
            v10 = vals[0, y_top_ind, z_bot_ind]
            v11 = vals[0, y_top_ind, z_top_ind]

        v0 = v00*(1.0-y_fac) + v10*y_fac
        v1 = v01*(1.0-y_fac) + v11*y_fac

        result_array[i] = v0*(1-z_fac) + v1*z_fac


# BISECT DOESNT WORK WITH OPENMP... YIELDS SEGFAULT??
@cython.boundscheck(False)
@cython.wraparound(False)
cdef int bisect(np.float64_t x, int nval, np.float64_t [:] arr) nogil:
    cdef int mid, top, bot
    top = nval - 1
    bot = 0

    while(top > bot + 1) :
        mid = floor((top-bot)/2)+bot
        if (x > arr[mid]) :
            bot = mid
        else :
            top = mid
    return bot



"""
Methods for describing parts of files to load.

This module provides generalized logic for getting parts of sequential data off disk.
It is for internal use. If you want to write a loader that supports partial loading,
it will make it a lot easier.

The steps for loading particle data are as follows

1. Set up a description of the particles you have on disk. This is a dictionary
   mapping a family type to a slice, e.g.

   .. code-block:: python

      on_disk = {pynbody.family.dm: slice(0,100), pynbody.family.gas: slice(100, 150)}

   describes a file with 100 dark matter particles followed by 50 gas particles.

2. Create a :class:`LoadControl` object.

   .. code-block:: python

      load_control = pynbody.chunk.LoadControl(on_disk, chunk_size, take)

   Here,

   * on_disk is the dictionary you set up in the last step.

   * chunk_size is the maximum number of particles you are willing to load off
     disk at once these will have to be stored in a temporary array, so you don't
     want it to be too large; but also reading a small number of large chunks is
     more efficient, so you don't want it to be too small. No careful experimentation
     has been done with this, but chunk_sizes of around 10000 seem to work OK.

   * take describes what to load in. Currently this is either ``None`` (= load the whole file)
     or a list of ids (= load the specified particles). However this may be expanded
     in future to a more comprehensive syntax. The idea is your code will not have to
     change when this happens, and will automatically support more advanced partial loading
     specifications.

3. Load your particle data. The :class:`LoadControl` object has an ``iterate`` method. This returns
   step-by-step instructions that take you through the file, specifying what to keep and what
   to throw away in a simple-to-use fashion. See the help for :func:`LoadControl.iterate` for details
   on how to implement this final step.

"""

from __future__ import annotations

import copy
from typing import TYPE_CHECKING, Callable, Iterable, Iterator

import numpy as np

from .. import util

if TYPE_CHECKING:
    from .. import family

class LoadControl:
    """LoadControl provides the logic required for partial loading.

    See the documentation for :mod:`pynbody.chunk` for more information."""

    def __init__(self, family_slice: dict[family.Family, slice], max_chunk: int, clauses: np.ndarray | None):
        """Initialize a LoadControl object.

        *Inputs:*

          *family_slice*: a dictionary of family slices describing the contiguous
            layout of families on disk

          *max_chunk*: the guaranteed maximum chunk of data to load in a single
            read operation. Larger values are likely more efficient, but also require
            bigger temporary buffers in your reader code.

          *clauses*: a description of the type of partial loading to implement. If None, all data is loaded.
            Otherwise, currently the only supported option is a numpy array of particle ids to load.
         """

        self._disk_family_slice = family_slice
        self._generate_family_order()

        # generate simulation-level ID list
        if hasattr(clauses, "__len__"):
            self._ids = np.asarray(clauses)
        else:
            self._ids = None  # no partial loading!

        self.generate_family_id_lists()
        self._generate_mem_slice()

        self.mem_num_particles = self.mem_family_slice[
            self._ordered_families[-1]].stop
        self.disk_num_particles = self._disk_family_slice[
            self._ordered_families[-1]].stop

        self._generate_chunks(max_chunk)

    @staticmethod
    def _scan_for_next_stop(ids, offset_start, id_maximum):
        from . import scan
        return scan.scan_for_next_stop(ids, offset_start, id_maximum)

    def generate_family_id_lists(self):
        if self._ids is None:
            self._family_ids = None
            return

        self._family_ids = {}
        offset = 0
        stop = 0
        for fam in self._ordered_families:
            sl = self._disk_family_slice[fam]
            self._family_ids[fam] = self._ids[
                (self._ids >= sl.start) * (self._ids < sl.stop)] - sl.start

    def _generate_family_order(self):
        famlist = []
        for fam, sl in self._disk_family_slice.items():
            famlist.append((fam, sl.start, sl.stop))

        famlist.sort(key=lambda x: x[1])
        famlist.sort(key=lambda x: x[2])
        self._ordered_families = [x[0] for x in famlist]

    def _generate_mem_slice(self):
        if self._ids is None:
            self.mem_family_slice = self._disk_family_slice
            return

        self.mem_family_slice = {}
        stop = 0
        for current_family in self._ordered_families:

            start = stop
            stop = stop + len(self._family_ids[current_family])
            self.mem_family_slice[current_family] = slice(start, stop)

    def _generate_null_chunks(self, max_chunk):
        """Generate internal chunk map in the special case that we are loading all data.

        See also :func:`_generate_chunks` for the general case.
        """

        self._family_chunks = {}

        for current_family in self._ordered_families:
            self._family_chunks[current_family] = []
            disk_sl = self._disk_family_slice[current_family]
            for i0 in range(0, disk_sl.stop - disk_sl.start, max_chunk):
                nread = min(disk_sl.stop - disk_sl.start - i0, max_chunk)
                buf_sl = slice(0, nread)
                mem_sl = slice(i0, i0 + nread)

                self._family_chunks[current_family].append(
                    (nread, buf_sl, mem_sl))

    def _generate_chunks(self, max_chunk):
        """Generate internal chunk map

        This must satisfy the requirements that:
          * maximum chunk size does not exceed *max_chunk*
          * chunks do not cross family boundaries

        """

        if self._ids is None:
            self._generate_null_chunks(max_chunk)
            return

        self._family_chunks = {}

        for current_family in self._ordered_families:
            self._family_chunks[current_family] = []
            disk_sl = self._disk_family_slice[current_family]
            ids = self._family_ids[current_family]
            i = 0

            disk_ptr = 0
            mem_ptr = 0

            while disk_ptr < disk_sl.stop - disk_sl.start:
                disk_ptr_end = disk_ptr + \
                    min(disk_sl.stop - disk_sl.start - disk_ptr, max_chunk - 1)
                j = self._scan_for_next_stop(ids, i, disk_ptr_end - 1)

                nread_disk = disk_ptr_end - disk_ptr

                assert (ids[i:j] < disk_ptr_end).all()

                if i != j:
                    mem_slice = slice(mem_ptr, mem_ptr + j - i)
                else:
                    mem_slice = None

                disk_mask = ids[i:j] - disk_ptr

                mem_ptr = mem_ptr + j - i
                i = j
                disk_ptr = disk_ptr_end

                self._family_chunks[current_family].append(
                    (nread_disk, disk_mask, mem_slice))

    def iterate_with_interrupts(self, families_on_disk: list[family.Family], families_in_memory: list[family.Family],
                                disk_interrupt_points: Iterable[int], disk_interrupt_fn: Callable,
                                multiskip: bool = False):
        """Yields instructions for loading an array with the specified families, breaking at specified file offsets

        Performs the same function as :func:`iterate` but additionally takes a list of exact file offsets
        *disk_interrupt_points* at which to interrupt the loading process and call a user-specified function
        *disk_interrupt_fn*.

        Parameters
        ----------

        disk_interrupt_points: Iterable
            List (or other iterable) of disk offsets at which to call the interrupt function, in ascending order

        disk_interrupt_fn: Callable
            Function which takes the file offset as an argument, and is called precisely at the point that the disk
            interrupt point is reached


        See :func:`iterate` for other parameters.

        """

        fpos = 0
        i = 0
        next_dip = disk_interrupt_points[i]
        for nread_disk, disk_slice, mem_slice in self.iterate(families_on_disk, families_in_memory, multiskip):
            while next_dip and fpos + nread_disk > next_dip:
                # an interrupt falls in the middle of our slice
                # work out what to read first
                len_pre = next_dip - fpos

                if disk_slice is not None:
                    d_slice_pre = util.concatenate_indexing(
                        slice(0, len_pre), disk_slice)
                    len_m_pre = util.indexing_length(d_slice_pre)
                    m_slice_pre = util.concatenate_indexing(
                        mem_slice, slice(0, len_m_pre))

                    # work out what to read second
                    len_post = nread_disk - len_pre
                    d_slice_post = copy.copy(util.concatenate_indexing(
                        disk_slice, slice(len_m_pre, util.indexing_length(mem_slice))))
                    # the copy above is necessary to ensure we don't end up inadvertently modifying
                    # list of offsets somewhere else

                    # that's the disk slice relative to having read the whole thing continuously.
                    # Offset to reflect what we've missed.

                    if isinstance(d_slice_post, slice):
                        d_slice_post = slice(
                            d_slice_post.start - len_pre, d_slice_post.stop - len_pre, d_slice_post.step)
                    else:
                        d_slice_post -= len_pre

                    len_m_post = util.indexing_length(d_slice_post)
                    m_slice_post = util.concatenate_indexing(
                        mem_slice, slice(len_m_pre, len_m_pre + len_m_post))

                    if util.indexing_length(d_slice_post) == 0:
                        d_slice_post = None
                        m_slice_post = None

                    if util.indexing_length(d_slice_pre) == 0:
                        d_slice_pre = None
                        m_slice_pre = None

                else:
                    d_slice_pre = None
                    m_slice_pre = None
                    len_post = nread_disk - len_pre
                    d_slice_post = None
                    m_slice_post = None

                yield len_pre, d_slice_pre, m_slice_pre
                fpos += len_pre
                disk_interrupt_fn(disk_interrupt_points[i])

                # prepare for next interrupt

                i += 1
                if len(disk_interrupt_points) > i:
                    next_dip = disk_interrupt_points[i]
                else:
                    next_dip = None

                # update 'input' to reflect post reading slice (may still need further
                # decomposition)
                nread_disk, disk_slice, mem_slice = len_post, d_slice_post, m_slice_post


            yield nread_disk, disk_slice, mem_slice
            fpos += nread_disk

    def iterate(self, families_on_disk: list[family.Family], families_in_memory: list[family.Family],
                multiskip: bool = False) -> Iterator[tuple[int, slice | None, slice | None]]:
        """Yields step-by-step instructions for partial-loading an array with the specified families.

        A typical read loop should be as follows:

        .. code-block:: python

            for readlen, buffer_index, memory_index in ctl.iterate(fams_on_disk, fams_in_mem) :
              data = read_entries(count=readlen)
              if memory_index is not None :
                target_array[memory_index] = data[buffer_index]

        Obviously this can be optimized, for instance to skip through
        file data when memory_index is None rather than read and discard it.

        Parameters
        ----------
        families_on_disk : list
            List of families for which the array exists on disk
        families_in_memory : list
            List of families for which we want to read the array into memory
        multiskip : bool
            If True, skip commands (i.e. entries with buffer_index=None)
            can have readlen greater than the block length

        Yields
        ------
        readlen : int
            Number of entries to read from disk
        buffer_index : slice | None
            Slice to read from the resulting buffer, or None if this particular read is to be ignored (skipped)
        memory_index : slice | None
            Slice to write into memory, or None if ``buffer_index`` is None

        """

        mem_offset = 0

        skip_accumulation = 0

        for current_family in self._ordered_families:
            if skip_accumulation > 0:
                yield skip_accumulation, None, None
                skip_accumulation = 0

            if current_family not in families_on_disk:
                assert current_family not in families_in_memory
            else:
                if current_family in families_in_memory:
                    for nread_disk, disk_mask, mem_slice in self._family_chunks[current_family]:
                        if mem_slice is None:
                            if multiskip:
                                skip_accumulation += nread_disk
                            else:
                                yield nread_disk, None, None
                        else:
                            if skip_accumulation > 0:
                                yield skip_accumulation, None, None
                                skip_accumulation = 0
                            mem_slice_offset = slice(
                                mem_slice.start + mem_offset, mem_slice.stop + mem_offset)
                            yield nread_disk, disk_mask, mem_slice_offset

                    mem_fs = self.mem_family_slice[current_family]
                    mem_offset += mem_fs.stop - mem_fs.start
                else:
                    for nread_disk, disk_mask, mem_slice in self._family_chunks[current_family]:
                        yield nread_disk, None, None



import pynbody
import pynbody.array as pyn_array
import pynbody.array.shared as shared

SA = pynbody.array.SimArray
import gc
import os
import signal
import sys
import time

import numpy as np
import numpy.testing as npt
import pytest


def test_pickle():
    import pickle
    x = SA([1, 2, 3, 4], units='kpc')
    assert str(x.units) == 'kpc'
    y = pickle.loads(pickle.dumps(x))
    assert y[3] == 4
    assert str(y.units) == 'kpc'

def test_issue255() :
    r = SA(0.5, 'au')
    assert isinstance(r**2, SA)
    assert float(r**2)==0.25
    assert str((r**2).units)=="au**2"


def test_return_types():

    x = SA([1, 2, 3, 4])
    y = SA([2, 3, 4, 5])

    assert type(x) is SA
    assert type(x ** 2) is SA
    assert type(x + y) is SA
    assert type(x * y) is SA
    assert type(x ** y) is SA
    assert type(2 ** x) is SA
    assert type(x + 2) is SA
    assert type(x[:2]) is SA

    x2d = SA([[1, 2, 3, 4], [5, 6, 7, 8]])

    assert type(x2d.sum(axis=1)) is SA


def test_unit_tracking():

    x = SA([1, 2, 3, 4])
    x.units = "kpc"

    y = SA([5, 6, 7, 8])
    y.units = "Mpc"

    assert abs((x * y).units.ratio("kpc Mpc") - 1.0) < 1.e-9

    assert ((x ** 2).units.ratio("kpc**2") - 1.0) < 1.e-9

    assert ((x / y).units.ratio("") - 1.e-3) < 1.e-12

    if hasattr(np.mean(x), 'units'):
        assert np.var(x).units == "kpc**2"
        assert np.std(x).units == "kpc"
        assert np.mean(x).units == "kpc"


def test_iop_units():
    x = SA([1, 2, 3, 4])
    x.units = 'kpc'

    y = SA([2, 3, 4, 5])
    y.units = 'km s^-1'

    z = SA([1000, 2000, 3000, 4000])
    z.units = 'm s^-1'

    assert repr(x) == "SimArray([1, 2, 3, 4], 'kpc')"

    with pytest.raises(ValueError):
        x += y

    x *= pynbody.units.Unit('K')

    assert x.units == 'K kpc'

    x.units = 'kpc'

    x *= y

    assert x.units == 'km s^-1 kpc'
    assert (x == [2, 6, 12, 20]).all()

    y += z
    assert y.units == 'km s^-1'

    assert (y == [3, 5, 7, 9]).all()


def test_iop_sanity():
    x = SA([1.0, 2.0, 3.0, 4.0])
    x_id = id(x)
    x += 1
    assert id(x) == x_id
    x -= 1
    assert id(x) == x_id
    x *= 2
    assert id(x) == x_id
    x /= 2
    assert id(x) == x_id


def test_unit_array_interaction():
    """Test for issue 113 and related"""
    x = pynbody.units.Unit('1 Mpc')
    y = SA(np.ones(10), 'kpc')
    npt.assert_allclose(x + y, SA([1.001] * 10, 'Mpc'))
    npt.assert_allclose(x - y, SA([0.999] * 10, 'Mpc'))

    assert (x + y).units == 'Mpc'

    npt.assert_allclose(y + x, SA([1.001] * 10, 'Mpc'))
    npt.assert_allclose(y - x, SA([-999.] * 10, 'kpc'))

def test_norm_units():
    x = SA(np.ones((10, 3) ), "kpc")
    result = np.linalg.norm(x, axis=1)
    npt.assert_allclose(result, np.ones(10) * np.sqrt(3), rtol=1.e-5)
    assert result.units == "kpc"

def test_dimensionful_comparison():
    # check that dimensionful units compare correctly
    # see issue 130
    a1 = SA(np.ones(2), 'kpc')
    a2 = SA(np.ones(2) * 2, 'pc')
    assert (a2 < a1).all()
    assert not (a2 > a1).any()
    a2 = SA(np.ones(2) * 1000, 'pc')
    assert (a1 == a2).all()
    assert (a2 <= a2).all()

    a2 = SA(np.ones(2), 'Msol')
    with pytest.raises(pynbody.units.UnitsException):
        a2 < a1

    a2 = SA(np.ones(2))
    with pytest.raises(pynbody.units.UnitsException):
        a2 < a1

    assert (a1 < pynbody.units.Unit("0.5 Mpc")).all()
    assert (a1 > pynbody.units.Unit("400 pc")).all()

    # now check with subarrays

    x = pynbody.new(10)
    x['a'] = SA(np.ones(10), 'kpc')
    x['b'] = SA(2 * np.ones(10), 'pc')

    y = x[[1, 2, 5]]

    assert (y['b'] < y['a']).all()
    assert not (y['b'] > y['a']).any()

def test_issue_485_1():
    s = pynbody.load("testdata/gadget2/test_g2_snap.1")
    stars = s.s
    indexed_arr = stars[1,2]
    np.testing.assert_almost_equal(np.sum(indexed_arr['vz'].in_units('km s^-1')), -20.13701057434082031250)
    np.testing.assert_almost_equal(np.std(indexed_arr['vz'].in_units('km s^-1')), 11.09318065643310546875)

def test_issue_485_2():
    # Adaptation of examples/vdisp.py
    s = pynbody.load("testdata/gadget2/test_g2_snap.1")

    stars = s.s
    rxyhist, rxybins = np.histogram(stars['rxy'], bins=20)
    rxyinds = np.digitize(stars['rxy'], rxybins)
    nrbins = len(np.unique(rxyinds))
    sigvz = np.zeros(nrbins)
    sigvr = np.zeros(nrbins)
    sigvt = np.zeros(nrbins)
    rxy = np.zeros(nrbins)

    assert len(np.unique(rxyinds)) == 3
    for i, ind in enumerate(np.unique(rxyinds)):
        bininds = np.where(rxyinds == ind)
        sigvz[i] = np.std(stars[bininds]['vz'].in_units('km s^-1'))
        sigvr[i] = np.std(stars[bininds]['vr'].in_units('km s^-1'))
        sigvt[i] = np.std(stars[bininds]['vt'].in_units('km s^-1'))
        rxy[i] = np.mean(stars[bininds]['rxy'].in_units('kpc'))


    np.testing.assert_allclose(sigvz, np.array([19.68325233, 29.49512482,  0.]), rtol=1e-6)
    np.testing.assert_allclose(sigvr, np.array([25.64306641, 26.01454544,  0.]), rtol=1e-6)
    np.testing.assert_allclose(sigvt, np.array([28.49997711, 18.84262276,  0.]), rtol=1e-6)
    np.testing.assert_allclose(rxy, np.array([1136892.125, 1606893.625, 1610494.75]), rtol=1e-6)

def _test_and_alter_shared_value(array_info):
    array = pynbody.array.shared._shared_array_reconstruct(array_info)
    assert (array[:] == np.arange(3)[: , np.newaxis] * np.arange(5)[np.newaxis, :]).all()
    array[:] = np.arange(3)[:, np.newaxis]

def test_shared_arrays():

    gc.collect() # this is to start with a clean slate, get rid of any shared arrays that might be hanging around
    baseline_num_shared_arrays = pyn_array.shared.get_num_shared_arrays_owned() # hopefully zero, but we can't guarantee that

    ar = pyn_array.array_factory((3, 5), dtype=np.float32, zeros=True, shared=True)

    assert ar.shape == (3,5)
    assert (ar == 0.0).all()

    ar[:] = np.arange(3)[: , np.newaxis] * np.arange(5)[np.newaxis, :]

    # now let's see if we can transfer it to another process:

    import multiprocessing as mp
    context = mp.get_context('spawn')
    p = context.Process(target=_test_and_alter_shared_value, args=(pyn_array.shared._shared_array_deconstruct(ar),))
    p.start()
    p.join()

    # check that the other process has successfully changed our value:
    assert (ar[:] ==  np.arange(3)[:, np.newaxis] ).all()

    assert pyn_array.shared.get_num_shared_arrays_owned() == 1 + baseline_num_shared_arrays

    ar2 = pyn_array.array_factory((3, 5), dtype=np.float32, zeros=True, shared=True)
    assert pyn_array.shared.get_num_shared_arrays_owned() == 2 + baseline_num_shared_arrays

    del ar, ar2
    gc.collect()

    assert pyn_array.shared.get_num_shared_arrays_owned() == baseline_num_shared_arrays

def test_shared_array_ownership():
    """Test that we can have two copies of a shared array in a process, but that only the 'owner' cleans up the memory"""

    import pynbody.array as pyn_array

    baseline_num_shared_arrays = pyn_array.shared.get_num_shared_arrays_owned()  # hopefully zero, but we can't guarantee that
    ar = pyn_array.array_factory((10,), int, True, True)
    assert pyn_array.shared.get_num_shared_arrays_owned() == 1 + baseline_num_shared_arrays

    array_info = pyn_array.shared._shared_array_deconstruct(ar)
    ar2 = pynbody.array.shared._shared_array_reconstruct(array_info)
    del ar2

    gc.collect()

    # shouldn't have been deleted!
    assert pyn_array.shared.get_num_shared_arrays_owned() == 1 + baseline_num_shared_arrays



@pytest.fixture
def clean_up_test_protection():
    import posix_ipc
    try:
        posix_ipc.unlink_shared_memory("pynbody-test-cleanup")
    except posix_ipc.ExistentialError:
        pass
    yield
    try:
        posix_ipc.unlink_shared_memory("pynbody-test-cleanup")
    except posix_ipc.ExistentialError:
        pass

def _test_shared_arrays_cleaned_on_exit():
    global ar
    ar = shared.make_shared_array((10,), dtype=np.int32, zeros=True, fname="pynbody-test-cleanup")
    # intentionally don't delete it, to see if it gets cleaned up on exit

def test_shared_arrays_cleaned_on_exit(clean_up_test_protection):
    _run_function_externally("_test_shared_arrays_cleaned_on_exit")

    _assert_shared_memory_cleaned_up()


def _test_shared_arrays_cleaned_on_terminate():
    # designed to run in a completely separate python process (i.e. not a subprocess of the test process)
    ar = shared.make_shared_array((10,), dtype=np.int32, zeros=True, fname="pynbody-test-cleanup")

    # send SIGTERM to ourselves:
    os.kill(os.getpid(), signal.SIGTERM)

    # wait to die...
    time.sleep(2.0)


def test_shared_arrays_cleaned_on_kill(clean_up_test_protection):
    stderr = _run_function_externally("_test_shared_arrays_cleaned_on_terminate")
    _assert_shared_memory_cleaned_up()


def _assert_shared_memory_cleaned_up():
    with pytest.raises(pynbody.array.shared.SharedArrayNotFound):
        _ = shared.make_shared_array((10,), dtype=np.int32, zeros=False,
                                     fname="pynbody-test-cleanup", create=False)


def _run_function_externally(function_name):
    pwd = os.path.dirname(__file__)
    python = sys.executable
    import subprocess
    process = subprocess.Popen([python, "-c",
                                f"from array_test import {function_name}; {function_name}()"]
                               , cwd=pwd)
    process.wait()



"""
Filters are used to define subsets of simulations, especially (but not exclusively) spatial sub-regions.

The basic idea is that a :class:`Filter` object stores the abstract definition of the subset, and then it can be
called with a simulation to return a boolean array indicating which particles are in the subset. The implementation
of filters is designed to be as efficient as possible, and in many cases the selection is done in C with OpenMP
parallelisation. Additionally, if a simulation has a :class:`pynbody.kdtree.KDTree` built (via
:meth:`pynbody.snapshot.simsnap.SimSnap.build_tree`), then the selection can be done using the KDTree, which is considerably
faster for very large simulations.

Filters can be combined using the logical operators `&`, `|` and `~` to create more complex selections.

For a friendly introduction, see :doc:`/tutorials/filters`.
"""
from __future__ import annotations

import pickle
from typing import TYPE_CHECKING

import numpy as np
from numpy.typing import ArrayLike

from .. import family, units
from . import geometry_selection

if TYPE_CHECKING:
    from .. import snapshot

class Filter:
    """Base class for all filters. Filters are callables that take simulations as input and return a boolean mask"""

    def where(self, sim: snapshot.SimSnap):
        """Return the indices of particles that are in the filter.

        This is a convenience method that is equivalent to np.where(f(sim)) but may be more efficient for some filters.
        """
        return np.where(self(sim))

    def __call__(self, sim: snapshot.SimSnap):
        """Return a boolean mask indicating which particles are in the filter."""
        return np.ones(len(sim), dtype=bool)

    def __and__(self, f2):
        return And(self, f2)

    def __invert__(self):
        return Not(self)

    def __or__(self, f2):
        return Or(self, f2)

    def __repr__(self):
        return "Filter()"

    def __hash__(self):
        return hash(pickle.dumps(self))

    def __eq__(self, other):
        if type(self) is not type(other):
            return False

        for k, v in self.__dict__.items():
            if k not in other.__dict__:
                return False
            else:
                equal = other.__dict__[k]==v
                if isinstance(equal, np.ndarray):
                    equal = equal.all()
                if not equal:
                    return False

        return True

    def _get_wrap_in_position_units(self, sim):
        """Helper method to get the simulation box wrap in units of the position array.

        If the boxsize is undefined, returns -1.0"""
        sim_ancestor = sim.ancestor
        if 'boxsize' in sim.properties:
            wrap = sim_ancestor.properties['boxsize']
            if units.is_unit_like(wrap):
                wrap = float(wrap.in_units(sim_ancestor['pos'].units,
                                           **sim_ancestor.conversion_context()))
        else:
            wrap = -1.0

        return wrap

    def cubic_cell_intersection(self, centroids):
        """Compute the intersection with cubic cells with the specified centroids.

        This is currently used by the swift loader to figure out which cells to load"""
        raise NotImplementedError("Cell intersections are not implemented for this filter")

    @classmethod
    def _get_boxsize_and_delta_from_centroids(self, centroids):
        """Helper for calculating cubic cell intersections"""
        deltax = (centroids[1] - centroids[0]).max()
        # check we are interpreting this right: the ((deltax+max-min)/deltax)^3
        # should be the number of cells
        ncells = len(centroids)
        maxpos = centroids.max()
        minpos = centroids.min()
        boxsize = (deltax + maxpos - minpos)
        ncells_from_geometry = np.round(boxsize / deltax) ** 3
        assert ncells == ncells_from_geometry, "Geometry of cells doesn't match expectations"
        return boxsize, deltax

class FamilyFilter(Filter):
    """A filter that selects particles based on their family."""

    def __init__(self, family_):
        self.family = family.get_family(family_, False)

    def __repr__(self):
        return "FamilyFilter("+self.family.name+")"

    def __call__(self, sim):
        slice_ = sim._get_family_slice(self.family)
        flags = np.zeros(len(sim), dtype=bool)
        flags[slice_] = True
        return flags

class And(Filter):
    """A filter that selects particles that are in both of two other filters.

    You can construct this filter conveniently using the ``&`` operator, i.e.

    >>> f = f1 & f2

    returns a filter that selects particles that are in both ``f1`` and ``f2``.
    """

    def __init__(self, f1, f2):
        self.f1 = f1
        self.f2 = f2

    def __call__(self, sim):
        return self.f1(sim) * self.f2(sim)

    def __repr__(self):
        return "(" + repr(self.f1) + " & " + repr(self.f2) + ")"

    def cubic_cell_intersection(self, centroids):
        return self.f1.cubic_cell_intersection(centroids) & \
               self.f2.cubic_cell_intersection(centroids)

class Or(Filter):
    """A filter that selects particles that are in either of two other filters.

    You can construct this filter conveniently using the ``|`` operator, i.e.

    >>> f = f1 | f2

    returns a filter that selects particles that are in either ``f1`` or ``f2``."""

    def __init__(self, f1, f2):
        self.f1 = f1
        self.f2 = f2

    def __call__(self, sim):
        return self.f1(sim) + self.f2(sim)

    def __repr__(self):
        return "(" + repr(self.f1) + " | " + repr(self.f2) + ")"

    def cubic_cell_intersection(self, centroids):
        return self.f1.cubic_cell_intersection(centroids) | \
            self.f2.cubic_cell_intersection(centroids)


class Not(Filter):
    """A filter that selects particles that are not in another filter.

    You can construct this filter conveniently using the ``~`` operator, i.e.

    >>> f = ~f1

    returns a filter that selects particles that are not in ``f1``.
    """

    def __init__(self, f):
        self.f = f

    def __call__(self, sim):
        x = self.f(sim)
        return np.logical_not(x)

    def __repr__(self):
        return "~" + repr(self.f)

    def cubic_cell_intersection(self, centroids):
        return ~self.f1.cubic_cell_intersection(centroids)


class Sphere(Filter):
    """
    A filter that selects particles within `radius` of the point `cen`.
    """

    def __init__(self, radius: float | str | units.UnitBase, cen: ArrayLike = (0, 0, 0)):
        """Create a sphere filter.

        Parameters
        ----------

        radius :
            The radius of the sphere. If a string, it is interpreted as a unit string.

        cen :
            The centre of the sphere. If a :class:`pynbody.snapshot.simsnap.SimArray`, units can be provided and
            will be correctly accounted for.
        """

        self.cen = np.asarray(cen)
        if self.cen.shape != (3,):
            raise ValueError("Centre must be length 3 array")

        if isinstance(radius, str):
            radius = units.Unit(radius)

        self.radius = radius

    def __call__(self, sim):
        with sim.immediate_mode:
            pos = sim['pos']

        cen, radius = self._get_cen_and_radius_as_float(pos)

        wrap = self._get_wrap_in_position_units(sim)

        return geometry_selection.selection(np.asarray(pos),'sphere',(cen[0], cen[1], cen[2], radius), wrap)

    def _get_cen_and_radius_as_float(self, pos):
        radius = self.radius
        cen = self.cen
        if units.has_units(cen):
            cen = cen.in_units(pos.units)
        if units.is_unit_like(radius):
            radius = float(radius.in_units(pos.units,
                                           **pos.conversion_context()))
        return cen, radius

    def where(self, sim):
        if hasattr(sim, "kdtree"):
            cen, radius = self._get_cen_and_radius_as_float(sim["pos"])
            return (np.sort(sim.kdtree.particles_in_sphere(cen, radius)),)
        else:
            return super().where(sim)

    def cubic_cell_intersection(self, centroids):
        boxsize, deltax = self._get_boxsize_and_delta_from_centroids(centroids)

        # the maximum offset from the cell centre to any corner:
        expand_distance = (deltax/2) * np.sqrt(3)

        return geometry_selection.selection(np.asarray(centroids), 'sphere',
                                           (self.cen[0], self.cen[1], self.cen[2], self.radius+expand_distance),
                                            boxsize)

    def __repr__(self):
        if units.is_unit(self.radius):
            return f"Sphere('{str(self.radius)}', {repr(self.cen)})"
        else:
            return f"Sphere({self.radius:.2e}, {repr(self.cen)})"


class Cuboid(Filter):
    """A filter that selects particles within a cuboid defined by two opposite corners."""

    def __init__(self, x1: float | str | units.UnitBase, y1: float | str | units.UnitBase = None,
                 z1: float | str | units.UnitBase = None, x2: float | str | units.UnitBase = None,
                 y2: float | str | units.UnitBase = None, z2: float | str | units.UnitBase = None):
        """Create a cuboid filter.

        If any of the cube coordinates ``x1``, ``y1``, ``z1``, ``x2``, ``y2``, ``z2`` are not specified they are
        determined as ``y1=x1``; ``z1=x1``; ``x2=-x1``; ``y2=-y1``; ``z2=-z1``.
        """
        x1, y1, z1, x2, y2, z2 = (
            units.Unit(x) if isinstance(x, str) else x for x in (x1, y1, z1, x2, y2, z2))
        if y1 is None:
            y1 = x1
        if z1 is None:
            z1 = x1
        if x2 is None:
            x2 = -x1
        if y2 is None:
            y2 = -y1
        if z2 is None:
            z2 = -z1
        self.x1, self.y1, self.z1, self.x2, self.y2, self.z2 = x1, y1, z1, x2, y2, z2

    def __call__(self, sim):
        wrap = self._get_wrap_in_position_units(sim)
        return self._get_mask(sim['pos'], self._get_boundaries(sim, wrap), wrap)

    def _get_mask(self, pos, boundaries, wrap):
        x1, y1, z1, x2, y2, z2 = boundaries
        return geometry_selection.selection(pos, 'cube', (x1, y1, z1, x2, y2, z2), wrap).view(dtype=bool)

    def _get_boundaries(self, sim, wrap=None):
        if wrap is None:
            wrap = self._get_wrap_in_position_units(sim)
        x1,y1,z1,x2,y2,z2 = (x.in_units(sim["pos"].units, **sim["pos"].conversion_context())
                            if units.is_unit_like(x) else x
                            for x in (self.x1, self.y1, self.z1, self.x2, self.y2, self.z2))
        if x2 < x1:
            x2 += wrap
        if y2 < y1:
            y2 += wrap
        if z2 < z1:
            z2 += wrap

        if x2 < x1 or y2 < y1 or z2 < z1:
            raise ValueError("Cuboid boundaries are not well defined")

        return x1, y1, z1, x2, y2, z2

    def _get_bounding_sphere(self, cuboid_boundaries):
        x1, y1, z1, x2, y2, z2 = cuboid_boundaries
        return ((x1 + x2) / 2, (y1 + y2) / 2, (z1 + z2) / 2), np.sqrt(
            (x2 - x1) ** 2 + (y2 - y1) ** 2 + (z2 - z1) ** 2) / 2

    def where(self, sim):

        if hasattr(sim, "kdtree"):
            # KDTree doesn't currently natively find cuboid regions, so we get the bounding sphere
            # and the search for the cuboid within that
            cuboid_boundaries = self._get_boundaries(sim)
            cen, radius = self._get_bounding_sphere(cuboid_boundaries)
            in_sphere_particles = sim.kdtree.particles_in_sphere(cen, radius)
            pos = sim["pos"][in_sphere_particles]
            wrap = self._get_wrap_in_position_units(sim)
            mask = self._get_mask(pos, cuboid_boundaries, wrap)
            return np.sort(in_sphere_particles[mask]),
        else:
            return super().where(sim)

    def cubic_cell_intersection(self, centroids):
        boxsize, deltax = self._get_boxsize_and_delta_from_centroids(centroids)
        shift = deltax/2
        boundaries = (self.x1 - shift, self.y1 - shift, self.z1 - shift,
                      self.x2 + shift, self.y2 + shift, self.z2 + shift)
        return self._get_mask(centroids, boundaries, boxsize)



    def __repr__(self):
        x1, y1, z1, x2, y2, z2 = ("'%s'" % str(x)
                                  if units.is_unit_like(x) else x
                                  for x in (self.x1, self.y1, self.z1, self.x2, self.y2, self.z2))
        return f"Cuboid({x1}, {y1}, {z1}, {x2}, {y2}, {z2})"


class Disc(Filter):
    """A filter that selects particles within a disc of specified extent and thickness."""

    def __init__(self, radius: float | str | units.UnitBase, height: float | str | units.UnitBase,
                 cen: ArrayLike = (0, 0, 0)):
        """Create a disc filter.

        In keeping with other parts of pynbody, the disc is defined in the x-y plane, with the z-axis being the
        symmetry axis. This is useful in conjunction with automated disc alignment e.g.
        :meth:`pynbody.analysis.angmom.sideon`.

        Parameters
        ----------

        radius :
            The radius of the disc (in the xy-plane). If a string, it is interpreted as a unit string.

        height :
            The thickness of the disc (in the z-direction). If a string, it is interpreted as a unit string.

        cen :
            The centre of the disc. If a :class:`pynbody.snapshot.simsnap.SimArray`, units can be provided and
            will be correctly accounted for.
        """
        self.cen = np.asarray(cen)
        if self.cen.shape != (3,):
            raise ValueError("Centre must be length 3 array")

        if isinstance(radius, str):
            radius = units.Unit(radius)

        if isinstance(height, str):
            height = units.Unit(height)

        self.radius = radius
        self.height = height

    def __call__(self, sim):
        radius = self.radius
        height = self.height

        if units.is_unit_like(radius):
            radius = float(
                radius.in_units(sim["pos"].units, **sim["pos"].conversion_context()))
        if units.is_unit_like(height):
            height = float(
                height.in_units(sim["pos"].units, **sim["pos"].conversion_context()))
        distance = (((sim["pos"] - self.cen)[:, :2]) ** 2).sum(axis=1)
        return (distance < radius ** 2) * (np.abs(sim["z"] - self.cen[2]) < height)

    def __repr__(self):
        radius = self.radius
        height = self.height

        radius, height = (
            ("'%s'" % str(x) if units.is_unit_like(x) else '%.2e' % x) for x in (radius, height))

        return f"Disc({radius}, {height}, {repr(self.cen)})"


class BandPass(Filter):
    """Selects particles in a bandpass of a named property"""

    def __init__(self, prop: str, min: float | str | units.UnitBase, max: float | str | units.UnitBase):
        """Create a bandpass filter.

        Parameters
        ----------

        prop :
            The name of the simulation array to filter on.

        min :
            The minimum value of the property. If a string, it is interpreted as a unit string.

        max :
            The maximum value of the property. If a string, it is interpreted as a unit string.
        """
        if isinstance(min, str):
            min = units.Unit(min)

        if isinstance(max, str):
            max = units.Unit(max)

        self._prop = prop
        self._min = min
        self._max = max

    def __call__(self, sim):
        min_ = self._min
        max_ = self._max
        prop = self._prop

        if units.is_unit_like(min_):
            min_ = float(
                min_.in_units(sim[prop].units, **sim.conversion_context()))
        if units.is_unit_like(max_):
            max_ = float(
                max_.in_units(sim[prop].units, **sim.conversion_context()))

        return ((sim[prop] > min_) * (sim[prop] < max_))

    def __repr__(self):
        min_, max_ = (("'%s'" % str(x) if units.is_unit_like(
            x) else '%.2e' % x) for x in (self._min, self._max))
        return f"BandPass('{self._prop}', {min_}, {max_})"


class HighPass(Filter):
    """Selects particles exceeding a specified threshold of a named property
    """

    def __init__(self, prop: str, min: float | str | units.UnitBase):
        """Create a high-pass filter.

        Parameters
        ----------

        prop :
            The name of the simulation array to filter on.

        min :
            The minimum value of the property. If a string, it is interpreted as a unit string.
        """
        if isinstance(min, str):
            min = units.Unit(min)

        self._prop = prop
        self._min = min

    def __call__(self, sim):
        min_ = self._min

        prop = self._prop

        if units.is_unit_like(min_):
            min_ = float(
                min_.in_units(sim[prop].units, **sim.conversion_context()))

        return (sim[prop] > min_)

    def __repr__(self):
        min = ("'%s'" % str(self._min) if units.is_unit_like(
            self._min) else '%.2e' % self._min)
        return f"HighPass('{self._prop}', {min})"


class LowPass(Filter):

    """Selects particles below a specified threshold of a named property
    """

    def __init__(self, prop: str, max: float | str | units.UnitBase):
        """Create a low-pass filter.

        Parameters
        ----------

        prop :
            The name of the simulation array to filter on.

        max :
            The maximum value of the property. If a string, it is interpreted as a unit string.
        """
        if isinstance(max, str):
            max = units.Unit(max)

        self._prop = prop
        self._max = max

    def __call__(self, sim):
        max_ = self._max

        prop = self._prop

        if units.is_unit_like(max_):
            max_ = float(
                max_.in_units(sim[prop].units, **sim.conversion_context()))

        return (sim[prop] < max_)

    def __repr__(self):
        max = ("'%s'" % str(self._max) if isinstance(
            self._max, units.UnitBase) else '%.2e' % self._max)
        return f"LowPass('{self._prop}', {max})"


class Annulus(And):
    """A filter that selects particles in between two spheres specified by radii `r1` and `r2` centered on `cen`."""

    def __init__(self, r1: float | str | units.UnitBase, r2: float | str | units.UnitBase, cen: ArrayLike = (0, 0, 0)):
        """Create an annulus filter.

        Parameters
        ----------

        r1 :
            The inner radius of the annulus. If a string, it is interpreted as a unit string.

        r2 :
            The outer radius of the annulus. If a string, it is interpreted as a unit string.

        cen :
            The centre of the annulus. If a :class:`pynbody.snapshot.simsnap.SimArray`, units can be provided and
            will be correctly accounted for.
        """
        super().__init__(~Sphere(r1, cen), Sphere(r2, cen))


class SolarNeighborhood(And):
    """A filter that selects particles in a disc between 2d radii `r1` and `r2` and thickness `height`.

    As for :class:`Disc`, the galaxy disc is defined in the x-y plane, with the z-axis being the symmetry axis.

    Default parameters are provided that are approximately the solar neighborhood (coarsely selected).
    """

    def __init__(self, r1: float | str | units.UnitBase = units.Unit("5 kpc"),
                 r2: float | str | units.UnitBase = units.Unit("10 kpc"),
                 height: float | str | units.UnitBase = units.Unit("2 kpc"),
                 cen: ArrayLike = (0, 0, 0)):
        """Create a solar neighborhood filter.

        Parameters
        ----------

        r1 :
            The inner radius of the disc. If a string, it is interpreted as a unit string.

        r2 :
            The outer radius of the disc. If a string, it is interpreted as a unit string.

        height :
            The thickness of the disc. If a string, it is interpreted as a unit string.

        cen :
            The centre of the disc. If a :class:`pynbody.snapshot.simsnap.SimArray`, units can be provided and
            will be correctly accounted for.
        """
        super().__init__(~Disc(r1, height, cen), Disc(r2, height, cen))



# distutils: language = c++

import numpy as np

cimport cython
cimport libc.math as cmath
cimport numpy as np
cimport openmp
from cython.parallel cimport prange
from libc.math cimport atan, pow
from libc.stdlib cimport free, malloc

from pynbody import config

ctypedef fused fused_float:
    np.float32_t
    np.float64_t

ctypedef fused fused_float_2:
    np.float32_t
    np.float64_t

ctypedef fused fused_float_3:
    np.float32_t
    np.float64_t

ctypedef fused fused_int:
    np.int32_t
    np.int64_t

ctypedef fused fused_int_2:
    np.int32_t
    np.int64_t

ctypedef fused fused_int_3:
    np.int32_t
    np.int64_t

ctypedef fused int_or_float:
    np.float32_t
    np.float64_t
    np.int32_t
    np.int64_t

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def find_boundaries(np.ndarray[fused_int, ndim=1] ordered) :
    """Given an ascending-ordered integer array starting at zero, return an array that gives the first
    element for each number. For example, calling with [0,0,0,1,2,2,3] should return [0,3,4,6]."""

    cdef np.ndarray[fused_int, ndim=1] boundaries = np.zeros(ordered[len(ordered)-1]+1,dtype=ordered.dtype) - 1
    cdef int n, size = len(ordered), current=ordered[0]-1

    with nogil :
        for n in range(size) :
            if ordered[n]>=0 and current<ordered[n] :
                current = ordered[n]
                boundaries[current] = n


    return boundaries

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def _grid_gen_from_slice(sl, int nx, int ny, int nz, np.ndarray[fused_float, ndim=2] pos):
    cdef float x,y,z
    cdef int i,n, start, stop, step

    start = sl.start
    stop = sl.stop

    if sl.step is None:
        step = 1
    else:
        step = sl.step

    with nogil:
        n = start
        i=0
        while n<stop :
            x=n%nx
            y=(n//nx)%ny
            z=(n//(nx*ny))%nz
            pos[i,0]=(<fused_float>(x)+0.5)/nx
            pos[i,1]=(<fused_float>(y)+0.5)/ny
            pos[i,2]=(<fused_float>(z)+0.5)/nz
            i+=1
            n+=step

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def _grid_gen_from_indices(np.ndarray[fused_int, ndim=1] ind, int nx, int ny, int nz, np.ndarray[fused_float, ndim=2] pos):
    cdef float x,y,z
    cdef int i,n_i,N=len(ind)


    with nogil:
        for i in range(N):
            n_i = ind[i]
            x=n_i%nx
            y=(n_i//nx)%ny
            z=(n_i//(nx*ny))%nz
            pos[i,0]=(<fused_float>(x)+0.5)/nx
            pos[i,1]=(<fused_float>(y)+0.5)/ny
            pos[i,2]=(<fused_float>(z)+0.5)/nz



@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def grid_gen(indices_or_slice,  nx,  ny,  nz, pos=None):
    """Generate the x,y,z grid coordinates in the interval (0,1) for the
    specified indices (relative to the start of a GrafIC file) or slice of the
    file. nx,ny,nz are the number of particles in each dimension (presumably
    the same for all sane cases, but the file format allows for different
    values). If *pos* is not None, copy the results into the array; otherwise
    create a new array for the results and return it."""

    from . import indexing_length

    if pos is None:
        pos = np.empty((indexing_length(indices_or_slice), 3),dtype=float)

    if isinstance(indices_or_slice, slice):
        _grid_gen_from_slice(indices_or_slice,nx,ny,nz,pos)
    else:
        _grid_gen_from_indices(np.asarray(indices_or_slice),nx,ny,nz,pos)

    return pos

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def sum(np.ndarray[fused_float, ndim=1] ar):
    """OpenMP summation algorithm equivalent to numpy.sum(ar)"""
    cdef fused_float v
    cdef long i
    cdef long N=len(ar)
    cdef int num_threads = config['number_of_threads']
    for i in prange(N,nogil=True,schedule='static',num_threads=num_threads):
        v+=ar[i]
    return v

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def sum_if_gt(np.ndarray[fused_float, ndim=1] ar,
                   np.ndarray[fused_float_2, ndim=1] cmp_ar,
                   double cmp_ar_val):
    """OpenMP summation algorithm equivalent to numpy.sum(ar*(cmp_ar>cmp_ar_val))"""
    cdef fused_float v
    cdef long i
    cdef long N=len(ar)
    assert len(cmp_ar)==len(ar)
    cdef int num_threads = config['number_of_threads']
    for i in prange(N,nogil=True,schedule='static',num_threads=num_threads):
        if cmp_ar[i]>cmp_ar_val:
           v+=(ar[i])
    return v

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def sum_if_lt(np.ndarray[fused_float, ndim=1] ar,
                   np.ndarray[fused_float_2, ndim=1] cmp_ar,
                   double cmp_ar_val):
    """OpenMP summation algorithm equivalent to numpy.sum(ar*(cmp_ar<cmp_ar_val))"""
    cdef fused_float v
    cdef long i
    cdef long N=len(ar)
    cdef int num_threads = config['number_of_threads']
    assert len(cmp_ar)==len(ar)
    for i in prange(N,nogil=True,schedule='static',num_threads=num_threads):
        v+=(ar[i])*(cmp_ar[i]<cmp_ar_val)
    return v

@cython.boundscheck(False)
@cython.wraparound(False)
cdef np.int64_t search(fused_int a, fused_int_2[:] B,
                       fused_int_3[:] sorter,
                       fused_int_3 ileft, fused_int_3 iright) nogil:
    cdef fused_int_2 b
    cdef fused_int_3 imid
    while ileft <= iright:
        imid = (ileft + iright) // 2
        b = B[sorter[imid]]
        if b < a:
            ileft = imid + 1
        elif b > a:
            iright = imid - 1
        else:
            return imid
    return -1

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef np.ndarray[ndim=1, dtype=fused_int] binary_search(
        fused_int[:] a, fused_int_2[:] b,
        np.ndarray[fused_int_3, ndim=1] sorter, int num_threads=-1):
    """Search elements of a in b, assuming a, b[sorter] are sorted in increasing order.

    Parameters
    ----------
    a : int array (N, )
    b : int array (M, )
    sorter : int array(M, )
        The input arrays

    num_threads : int, optional
        If greater than zero, use that many parallel threads.


    Returns
    -------
    indices : array(N, )
        An array such that b[indices] == a, for elements found in b.
        Elements of a that cannot be found have the index M.

    Notes
    -----
    The code does *not* check that a and b[sorted] are effectively sorted in increasing order.

    This functions can be used in place of np.searchsorted (note however that the order of the argument differ).
    The algorithm performs particularly well when a is much smaller than b and can be found at close locations in b.

    Algorithm
    ---------
    The algorithm implemented is a binary search algorithm of the elements of a into b. The algorithm consumes elements
    of a from both ends. Since a is sorted, the index of a[0] and a[-1] give boundaries to look for the next elements
    of a in b, resulting in the next binary search being faster.
    If the elements of a are almost contiguous in b, the algorithm scales as N log(N), with N = len(a) instead of
    N log(M) with M = len(b).

    Note that the algorithm performs similarly to np.searchsorted when N~M.
    """

    cdef int Na = len(a), Nb = len(b)

    cdef int ileft=0, iright=Nb - 1
    cdef int i, ii, j, pivot
    cdef int index

    cdef fused_int_3[:] indices = np.empty(Na, dtype=sorter.dtype)

    # HACK: prevent cython from complaining about "buffer source array is read-only"
    cdef bint write_flag = sorter.flags['WRITEABLE']
    sorter.setflags(write=True)

    cdef fused_int_3[:] sorter_mview = sorter

    cdef int ichunk, chunk_size, Nchunk = config['number_of_threads'], this_chunk

    if num_threads > 0:
        Nchunk = num_threads
    openmp.omp_set_num_threads(Nchunk)

    if Na % Nchunk == 0:
        chunk_size = Na // Nchunk
    else:
        chunk_size = Na // Nchunk + 1

    for ichunk in prange(Nchunk, nogil=True, chunksize=1, schedule='static', num_threads=Nchunk):
        ileft = 0
        iright = Nb - 1
        this_chunk = min(chunk_size, Na-ichunk*chunk_size)
        pivot = (this_chunk + 1) // 2

        for ii in range(pivot):
            i = ichunk * chunk_size + ii
            j = ichunk * chunk_size + this_chunk - 1 - ii

            index = search(a[i], b, sorter_mview, ileft, iright)
            if index > -1:
                ileft = index
                indices[i] = sorter_mview[index]
            else:
                indices[i] = Nb


            if j > i:
                index = search(a[j], b, sorter_mview, ileft, iright)
                if index > -1:
                    iright = index
                    indices[j] = sorter_mview[index]
                else:
                    indices[j] = Nb

    # Restore write flag on sorter array
    sorter.setflags(write=write_flag)
    return np.asarray(indices)

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef int is_sorted(int_or_float[:] A):
    """Check whether input is sorted in ascending order.

    Arguments
    ---------
    A : array

    Returns
    -------
    ret : int
        +1 if A is in ascending order
        -1 if A is in descending order
        0 otherwise
    """
    cdef int Na = len(A), i, i0
    cdef int ret = 0

    # Special case for single-valued arrays
    if Na <= 1:
        return 1

    # Iterate until finding two consecutive non-null elements
    i0 = 1
    while i0 < Na:
        if A[i0] != A[0]:
            break
        i0 += 1

    # Special case if array is constant
    if i0 == Na:
        return 1

    if A[i0] >= A[0]:
        for i in range(i0, Na):
            if A[i] < A[i-1]:
                return 0
        return 1
    else:
        for i in range(i0, Na):
            if A[i] > A[i-1]:
                return 0
        return -1

__all__ = ['grid_gen','find_boundaries', 'sum', 'sum_if_gt', 'sum_if_lt',
           'binary_search', 'is_sorted']



"""
Implements classes and functions for handling RAMSES files.

Note that AMR cells are loaded as particles, although this works surprisingly well for most analysis purposes.
In particular SPH image generation routines automatically switch on an appropriate de-noising scheme to provide
effective interpolation between cells.

"""

import csv
import logging
import os
import re
import warnings
from pathlib import Path

import numpy as np

import pynbody.array.shared

from .. import array, config_parser, family, units
from ..analysis.cosmology import age
from ..extern.cython_fortran_utils import FortranFile
from . import SimSnap, namemapper

logger = logging.getLogger('pynbody.snapshot.ramses')

multiprocess_num = int(config_parser.get('ramses', "parallel-read"))
multiprocess = (multiprocess_num > 1)

if multiprocess:
    import multiprocessing

    remote_exec = pynbody.array.shared.shared_array_remote
    remote_map = pynbody.array.shared.remote_map
else:
    def remote_exec(fn):
        return fn

    def remote_map(*args, **kwargs):
        return list(map(*args[1:], **kwargs))


_float_type = 'd'
_int_type = 'i'

def _timestep_id(basename):
    try:
        return re.findall("output_([0-9]*)/*$", str(basename))[0]
    except IndexError:
        return None


def _cpu_id(i):
    return str(i).rjust(5, "0")


@remote_exec
def _cpui_count_particles_with_implicit_families(filename, distinguisher_field, distinguisher_type):

    with FortranFile(filename) as f:
        f.seek(0, 2)
        eof_fpos = f.tell()
        f.seek(0, 0)
        header = f.read_attrs(ramses_particle_header)
        npart_this = header['npart']
        f.skip(distinguisher_field)
        # Avoid end-of-file issues
        if f.tell() == eof_fpos:
            data = np.array([])
        else:
            data = f.read_vector(distinguisher_type)

        if len(data)>0:
            my_mask = np.array((data != 0), dtype=np.int8) # -> 0 for dm, 1 for star
        else:
            my_mask = np.zeros(npart_this, dtype=np.int8)
        nstar_this = (data != 0).sum()
        return npart_this, nstar_this, my_mask

@remote_exec
def _cpui_count_particles_with_explicit_families(filename, family_field, family_type):
    assert np.issubdtype(family_type, np.int8)
    counts_array = np.zeros(256,dtype=np.int64)
    with FortranFile(filename) as f:
        header = f.read_attrs(ramses_particle_header)
        npart_this = header['npart']

        f.skip(family_field)
        my_mask = f.read_vector(family_type)

        unique_mask_ids, counts = np.unique(my_mask, return_counts=True)
        counts_array[unique_mask_ids]=counts

        assert sum(counts)==npart_this

        return counts_array, my_mask

@remote_exec
def _cpui_load_particle_block(filename, arrays, offset, first_index, type_, family_mask):
    with FortranFile(filename) as f:
        _header = f.read_attrs(ramses_particle_header)
        f.skip(offset)
        data = f.read_vector(type_)
        for fam_id, ar in enumerate(arrays):
            data_this_family = data[family_mask == fam_id]
            ind0 = first_index[fam_id]
            ind1 = ind0 + len(data_this_family)
            ar[ind0:ind1] = data_this_family


def _cpui_level_iterator(cpu, amr_filename, bisection_order, maxlevel, ndim):
    with FortranFile(amr_filename) as f:
        header = f.read_attrs(ramses_amr_header)
        f.skip(13)

        n_per_level = f.read_vector(_int_type).reshape((header['nlevelmax'], header['ncpu']))
        f.skip(1)
        if header['nboundary'] > 0:
            f.skip(2)
            n_per_level_boundary = f.read_vector(_int_type).reshape((header['nlevelmax'], header['nboundary']))

        f.skip(2)
        if bisection_order:
            f.skip(5)
        else:
            f.skip(1)
        f.skip(3)

        offset = np.array(header['ng'], dtype=_float_type) / 2
        offset -= 0.5

        coords = np.zeros(3, dtype=_float_type)

        for level in range(maxlevel or header['nlevelmax']):

            # loop through those CPUs with grid data (includes ghost regions)
            for cpuf in 1 + np.where(n_per_level[level, :] != 0)[0]:
                # print "CPU=",cpu,"CPU on
                # disk=",cpuf,"npl=",n_per_level[level,cpuf-1]

                if cpuf == cpu:

                    # this is the data we want
                    f.skip(3)  # grid, next, prev index

                    # store the coordinates in temporary arrays. We only want
                    # to copy it if the cell is not refined
                    coords = [
                        f.read_vector(_float_type) for ar in range(ndim)]

                    # stick on zeros if we're in less than 3D
                    coords += [np.zeros_like(coords[0]) for ar in range(3 - ndim)]

                    f.skip(1  # father index
                                 + 2 * ndim  # nbor index
                                 # son index,cpumap,refinement map
                                 + 2 * (2 ** ndim)
                                 )

                    refine = np.array(
                        [f.read_vector(_int_type) for i in range(2 ** ndim)])

                    if(level+1 == maxlevel or level+1==header['nlevelmax']):
                        refine[:] = 0

                    coords[0] -= offset[0]
                    coords[1] -= offset[1]
                    coords[2] -= offset[2]
                    # x0-=offset[0]; y0-=offset[1]; z0-=offset[2]

                    yield coords, refine, cpuf, level

                else:

                    # skip ghost regions from other CPUs
                    f.skip(3 + ndim + 1 + 2 * ndim + 3 * 2 ** ndim)

            if header['nboundary'] > 0:
                for boundaryf in np.where(n_per_level_boundary[level, :] != 0)[0]:

                    f.skip(3 + ndim + 1 + 2 * ndim + 3 * 2 ** ndim)


@remote_exec
def _cpui_count_gas_cells(level_iterator_args):
    ncell = 0
    for coords, refine, cpu, level in _cpui_level_iterator(*level_iterator_args):
        ncell += (refine == 0).sum()
    return ncell


@remote_exec
def _cpui_load_gas_pos(pos_array, smooth_array, ndim, boxlen, i0, level_iterator_args):
    dims = [pos_array[:, i] for i in range(ndim)]
    subgrid_index = np.arange(2 ** ndim)[:, np.newaxis]
    subgrid_z = np.floor((subgrid_index) / 4)
    subgrid_y = np.floor((subgrid_index - 4 * subgrid_z) / 2)
    subgrid_x = np.floor(subgrid_index - 2 * subgrid_y - 4 * subgrid_z)
    subgrid_x -= 0.5
    subgrid_y -= 0.5
    subgrid_z -= 0.5

    for (x0, y0, z0), refine, cpu, level in _cpui_level_iterator(*level_iterator_args):
        dx = boxlen * 0.5 ** (level + 1)

        x0 = boxlen * x0 + dx * subgrid_x
        y0 = boxlen * y0 + dx * subgrid_y
        z0 = boxlen * z0 + dx * subgrid_z

        mark = np.where(refine == 0)

        i1 = i0 + len(mark[0])
        for q, d in zip(dims, [x0, y0, z0][:ndim]):
            q[i0:i1] = d[mark]

        smooth_array[i0:i1] = dx
        i0 = i1

_gv_load_hydro = 0
_gv_load_gravity = 1
_gv_load_rt = 2


@remote_exec
def _cpui_load_gas_vars(dims, maxlevel, ndim, filename, cpu, lia, i1,
                        mode=_gv_load_hydro):

    logger.info("Loading data from CPU %d", cpu)

    nvar = len(dims)
    grid_info_iter = _cpui_level_iterator(*lia)

    with FortranFile(filename) as f:
        exact_nvar = False
        if mode is _gv_load_hydro:
            header = f.read_attrs(ramses_hydro_header)
            nvar_file = header['nvarh']
        elif mode is _gv_load_gravity:
            header = f.read_attrs(ramses_grav_header)
            nvar_file = 4
        elif mode is _gv_load_rt:
            header = f.read_attrs(ramses_rt_header)
            nvar_file = header['nrtvar']
            exact_nvar = True
        else:
            raise ValueError("Unknown RAMSES load mode")

        if nvar_file != nvar and exact_nvar:
            raise ValueError("Wrong number of variables in RAMSES dump")
        elif nvar_file < nvar:
            warnings.warn("Fewer hydro variables are in this RAMSES dump than are defined in config.ini (expected %d, got %d in file)" % (
                nvar, nvar_file), RuntimeWarning)
            nvar = nvar_file
            dims = dims[:nvar]
        elif nvar_file > nvar:
            warnings.warn("More hydro variables (%d) are in this RAMSES dump than are defined in config.ini (%d)" % (
                nvar_file, nvar), RuntimeWarning)

        for level in range(maxlevel or header['nlevelmax']):

            for cpuf in range(1, header['ncpu'] + 1):
                flevel = f.read_int()
                ncache = f.read_int()
                assert flevel - 1 == level

                if ncache > 0:
                    if cpuf == cpu:

                        coords, refine, gi_cpu, gi_level = next(grid_info_iter)
                        mark = np.where(refine == 0)

                        assert gi_level == level
                        assert gi_cpu == cpu

                    if cpuf == cpu and len(mark[0]) > 0:
                        for icel in range(2 ** ndim):
                            i0 = i1
                            i1 = i0 + (refine[icel] == 0).sum()
                            for ar in dims:
                                ar[i0:i1] = f.read_vector(
                                    _float_type)[(refine[icel] == 0)]

                            f.skip(nvar_file - nvar)

                    else:
                        f.skip((2 ** ndim) * nvar_file)

            for boundary in range(header['nboundary']):
                flevel = f.read_int()
                ncache = f.read_int()
                if ncache > 0:
                    f.skip((2 ** ndim) * nvar_file)


ramses_particle_header = (
    ('ncpu', 1, 'i'),
    ('ndim', 1, 'i'),
    ('npart', 1, 'i'),
    ('randseed', -1, 'i'),
    ('nstar', 1, 'i'),
    ('mstar', 1, 'd'),
    ('mstar_lost', 1, 'd'),
    ('nsink', 1, 'i')
)

ramses_amr_header = (
    ('ncpu', 1, 'i'),
    ('ndim', 1, 'i'),
    ('ng', 3, 'i'),
    ('nlevelmax', 1, 'i'),
    ('ngridmax', 1, 'i'),
    ('nboundary', 1, 'i'),
    ('ngrid', 1, 'i'),
    ('boxlen', 1, 'd')
)

ramses_hydro_header = (
    ('ncpu', 1, 'i'),
    ('nvarh', 1, 'i'),
    ('ndim', 1, 'i'),
    ('nlevelmax', 1, 'i'),
    ('nboundary', 1, 'i'),
    ('gamma', 1, 'd')
)

ramses_grav_header = (
    ('ncpu', 1, 'i'),
    ('ndim', 1, 'i'),
    ('nlevelmax', 1, 'i'),
    ('nboundary', 1, 'i')
)

ramses_rt_header = (
    ('ncpu', 1, 'i'),
    ('nrtvar', 1, 'i'),
    ('ndim', 1, 'i'),
    ('nlevelmax', 1, 'i'),
    ('nboundary', 1, 'i'),
    ('gamma', 1, 'd')
)

TYPE_MAP = {'i4': 'i',
            'i8': 'l',
            'f4': 'f',
            'f8': 'd'}
particle_blocks = list(map(
    str.strip, config_parser.get('ramses', "particle-blocks").split(",")))
particle_format = [TYPE_MAP[str.strip(e)] for e in config_parser.get('ramses', "particle-format").split(",")]

hydro_blocks = [_.strip() for _ in config_parser.get('ramses', "hydro-blocks").split(",")]
grav_blocks = [_.strip() for _ in config_parser.get('ramses', "gravity-blocks").split(",")]
rt_blocks = [_.strip() for _ in config_parser.get('ramses', 'rt-blocks', raw=True).split(",")]

particle_distinguisher = [_.strip() for _ in config_parser.get('ramses', 'particle-distinguisher').split(",")]
positive_typemap = [family.get_family(str.strip(x)) for x in config_parser.get('ramses', 'type-mapping-positive').split(",")]

negative_typemap = [family.get_family(str.strip(x)) for x in config_parser.get('ramses', 'type-mapping-negative').split(",")]

def _read_descriptor(fname):
    """Read a RAMSES file descriptor and return a list of the variable names."""
    description = []
    name_mapping = namemapper.AdaptiveNameMapper('ramses-name-mapping')
    with open(fname) as fd:
        if fd.readline() != "# version:  1\n":
            raise OSError("Wrong file format")
        fd.readline()  # ivar, variable_name, variable_type
        for line in fd.readlines():
            i, name, dtype = (_.strip() for _ in line.split(","))

            description.append(name_mapping(name, reverse=True))
    return description


class RamsesSnap(SimSnap):
    """Implements loading of Ramses snapshots.

    Note that AMR cells are loaded as particles, although this works surprisingly well for most analysis purposes.
    In particular SPH image generation routines automatically switch on an appropriate de-noising scheme to provide
    effective interpolation between cells.
    """
    reader_pool = None

    def __init__(self, dirname, cpus=None, maxlevel=None, with_gas=True, force_gas=False, times_are_proper=None):
        """
        Initialize a RamsesSnap.

        Parameters
        ----------
        dirname : str
            Directory containing the RAMSES output
        cpus : list of int, optional
            A list of the CPU IDs to load. If not set, load all
            CPU's data.
        maxlevel : int, optional
            The maximum refinement level to load. If not set, the
            deepest level is loaded.
        with_gas : bool, optional
            If False, never load any gas cells (particles only).
            Default is True
        force_gas : bool, optional
            If True, load the AMR cells as "gas particles" even if
            they don't actually contain gas in the run. Default is
            False.
        times_are_proper : bool, optional
            If True, the times in the output are assumed to be proper
            times. If False, they are assumed to be conformal. If
            None (default), assume proper for non-cosmological simulations
            and conformal for cosmological ones.
        """

        global config
        super().__init__()

        self.__setup_parallel_reading()

        if isinstance(dirname, Path):
            dirname = str(dirname)

        self._timestep_id = _timestep_id(dirname)
        self._filename = dirname
        if os.path.isdir(dirname):
            self._dirname = dirname
        else:
            self._dirname = os.path.split(dirname)[0]
        self._translate_array_name = namemapper.AdaptiveNameMapper('ramses-name-mapping')

        self._load_sink_data_to_temporary_store()
        self._load_infofile()
        self._load_namelistfile()
        self._setup_particle_descriptor()

        assert self._info['ndim'] <= 3
        if self._info['ndim'] < 3:
            warnings.warn(
                "Snapshots with less than three dimensions are supported only experimentally", RuntimeWarning)

        self._ndim = self._info['ndim']
        self.ncpu = self._info['ncpu']
        if cpus is not None:
            self._cpus = cpus
        else:
            self._cpus = list(range(1, self.ncpu + 1))
        self._maxlevel = maxlevel

        type_map = self._count_particles()

        has_gas = os.path.exists(
            self._hydro_filename(1)
        ) or force_gas

        if not with_gas:
            has_gas = False

        if times_are_proper is not None:
            self.times_are_proper = times_are_proper

        ngas = self._count_gas_cells() if has_gas else 0

        if ngas>0:
            type_map[family.gas] = ngas

        count = 0
        for fam in type_map:
            self._family_slice[fam] = slice(count, count+type_map[fam])
            count+=type_map[fam]

        self._num_particles = count
        self._load_fluid_descriptors()
        self._load_rt_infofile()
        self._decorate()
        self._transfer_sink_data_to_family_array()

    def __setup_parallel_reading(self):
        if multiprocess:
            self._shared_arrays = True
            if (RamsesSnap.reader_pool is None):
                RamsesSnap.reader_pool = multiprocessing.Pool(multiprocess_num)

    def _load_fluid_descriptors(self):
        types = ["hydro", "grav"]
        default_blocks = [hydro_blocks, grav_blocks]
        descriptors_fnames = [os.path.join(self._filename, f"{ftype}_file_descriptor.txt") for ftype in types]

        for desc_type, default_block, descriptor_fname in zip(types, default_blocks, descriptors_fnames):
            try:
                block = _read_descriptor(descriptor_fname)
            except (FileNotFoundError, OSError):
                block = default_block

            setattr(self, f"_{desc_type}_blocks", block)

    def _load_rt_infofile(self):
        self._rt_blocks = []
        self._rt_blocks_3d = set()
        try:
            with open(os.path.join(self._filename, f"info_rt_{self._timestep_id}.txt")) as f:
                lines = f.readlines()
        except OSError:
            return

        self._load_info_from_specified_file(lines)

        for group in range(self._info['nGroups']):
            for block in rt_blocks:
                self._rt_blocks.append(block%group)

        self._rt_unit = self._info['unit_pf']*units.Unit("cm^-2 s^-1")

        for block in self._rt_blocks:
            self._rt_blocks_3d.add(self._array_name_1D_to_ND(block) or block)

    def _load_info_from_specified_file(self, lines):
        for line in lines:
            if '=' in line:
                name, val = list(map(str.strip, line.split('=')))
                try:
                    if '.' in val:
                        self._info[name] = float(val)
                    else:
                        self._info[name] = int(val)
                except ValueError:
                    self._info[name] = val

    def _load_infofile(self):
        self._info = {}
        info_fname = os.path.join(self._filename, f"info_{self._timestep_id}.txt")
        header_fname = os.path.join(self._filename, f"header_{self._timestep_id}.txt")
        with open(info_fname) as f:
            self._load_info_from_specified_file(f)
        try:
            # most of this file is unhelpful, but depending on the ramses
            # version, there may be information on the particle fields present
            with open(header_fname) as f:
                for line in f:
                    if "level" in line:
                        self._info['particle-blocks'] = line.split()
        except OSError:
            warnings.warn(
                "No header file found -- no particle block information available")

    def _load_namelist_from_specified_file(self, f):
        for line in f:
            line = line.split("!")[0]  # remove fortran comments
            if '=' in line:
                name, val = map(str.strip, line.split('='))
                if val == ".true.":
                    self._namelist[name] = True
                elif val == ".false.":
                    self._namelist[name] = False
                else:
                    try:
                        if '.' in val:
                            self._namelist[name] = float(val)
                        else:
                            self._namelist[name] = int(val)
                    except ValueError:
                        self._namelist[name] = val

    def _load_namelistfile(self):
        self._namelist = {}

        namelist_file = os.path.join(self._filename, "namelist.txt")

        if os.path.exists(namelist_file):
            with open(namelist_file) as f:
                try:
                    self._load_namelist_from_specified_file(f)
                except ValueError:
                    warnings.warn("Namelist found but unable to read.")

        else:
            warnings.warn("No namelist file found.")

    def _setup_particle_descriptor(self):
        try:
            self._load_particle_descriptor()
        except OSError:
            self._guess_particle_descriptor()
        self._has_explicit_particle_families = 'family' in self._particle_blocks

    def _load_particle_descriptor(self):
        with open(os.path.join(self._filename, "part_file_descriptor.txt")) as f:
            self._particle_blocks = []
            self._particle_types = []
            for line in f:
                if not line.startswith("#"):
                    ivar, name, dtype = list(map(str.strip,line.split(",")))
                    self._particle_blocks.append(self._translate_array_name(name, reverse=True))
                    self._particle_types.append(dtype)
            self._particle_blocks_are_explictly_known = True


    def _guess_particle_descriptor(self):
        # determine whether we have explicit information about
        # what particle blocks are present
        self._particle_blocks_are_explictly_known = False

        self._particle_types = particle_format
        if 'particle-blocks' in self._info:
            self._particle_blocks_are_explictly_known = True
            self._particle_blocks = self._info['particle-blocks']
            self._particle_blocks = ['x', 'y', 'z', 'vx', 'vy', 'vz'] + self._particle_blocks[2:]
        else:
            self._particle_blocks = particle_blocks

        if len(self._particle_types) < len(self._particle_blocks):
            warnings.warn("Some fields do not have format configured - assuming they are doubles", RuntimeWarning)
            type_ = 'd'
            self._particle_types += [type_] * (len(self._particle_blocks) - len(self._particle_types))

    def _particle_filename(self, cpu_id):
        return os.path.join(self._filename, f"part_{self._timestep_id}.out{_cpu_id(cpu_id)}")

    def _amr_filename(self, cpu_id):
        return os.path.join(self._filename, f"amr_{self._timestep_id}.out{_cpu_id(cpu_id)}")

    def _hydro_filename(self, cpu_id):
        return os.path.join(self._filename, f"hydro_{self._timestep_id}.out{_cpu_id(cpu_id)}")

    def _grav_filename(self, cpu_id):
        return os.path.join(self._filename, f"grav_{self._timestep_id}.out{_cpu_id(cpu_id)}")

    def _rt_filename(self, cpu_id):
        return os.path.join(self._filename, f"rt_{self._timestep_id}.out{_cpu_id(cpu_id)}")

    def _sink_filename(self):
        return os.path.join(self._filename, f"sink_{self._timestep_id}.csv")

    def _count_particles(self):
        if self._has_explicit_particle_families:
            return self._count_particles_using_explicit_families()
        else:
            ndm, nstar = self._count_particles_using_implicit_families()
            return {family.dm: ndm, family.star: nstar}

    def _has_particle_file(self):
        """Check whether the output has a particle file available"""
        if len(self._cpus)>0 :
            return os.path.exists(self._particle_filename(self._cpus[0]))
        else:
            return False

    def _count_particles_using_explicit_families(self):
        """Returns an ordered dictionary of family types based on the new explicit RAMSES particle file format"""
        if not self._has_particle_file():
            return {}
        family_block = self._particle_blocks.index('family')
        family_dtype = self._particle_types[family_block]
        self._particle_family_ids_on_disk = []
        self._particle_file_start_indices = []
        results = remote_map(self.reader_pool,
                             _cpui_count_particles_with_explicit_families,
                             [self._particle_filename(i) for i in self._cpus],
                             [family_block] * len(self._cpus),
                             [family_dtype] * len(self._cpus))

        aggregate_counts = np.zeros(256,dtype=np.int64)
        for counts, family_ids in results:
            aggregate_counts+=counts
            self._particle_family_ids_on_disk.append(family_ids)

        # The above family IDs are defined according to ramses' own internal system. We now need
        # to map them to pynbody family types

        nonzero_families = np.nonzero(aggregate_counts)[0]

        # map such that negative integers are correctly represented
        ramses_id_to_internal_id = np.zeros(256, dtype=np.uint8)
        internal_id_to_family = []
        aggregate_counts_remapped = np.zeros(256, dtype=np.int64)

        for ramses_family_id in nonzero_families:
            if ramses_family_id>128 or ramses_family_id == 0:
                neg_offset = (256 - ramses_family_id) % 256
                if neg_offset>len(negative_typemap):
                    pynbody_family = negative_typemap[-1]
                else:
                    pynbody_family = negative_typemap[neg_offset]
            else:
                if ramses_family_id>len(positive_typemap):
                    pynbody_family = positive_typemap[-1]
                else:
                    pynbody_family = positive_typemap[ramses_family_id-1]
            if pynbody_family in internal_id_to_family:
                internal_id = internal_id_to_family.index(pynbody_family)
            else:
                internal_id = len(internal_id_to_family)
                internal_id_to_family.append(pynbody_family)
            ramses_id_to_internal_id[ramses_family_id] = internal_id
            aggregate_counts_remapped[internal_id]+=aggregate_counts[ramses_family_id]

        # perform the remapping for our stored particle identifiers
        for fid in self._particle_family_ids_on_disk:
            fid[:] = ramses_id_to_internal_id[fid]

        return_d = {}
        self._particle_file_start_indices = [ [] for x in results]
        for internal_family_id in range(256):
            if aggregate_counts_remapped[internal_family_id]>0:
                fam = internal_id_to_family[internal_family_id]
                count = aggregate_counts_remapped[internal_family_id]
                return_d[fam] = count
                startpoint = 0
                for i,fid in enumerate(self._particle_family_ids_on_disk):
                    self._particle_file_start_indices[i].append(startpoint)
                    startpoint+=(fid==internal_family_id).sum()
                assert startpoint==count

        n_sinks = self._count_sink_particles()

        if n_sinks>0:
            return_d[self._sink_family] = n_sinks

        return return_d


    def _load_sink_data_to_temporary_store(self):
        if not os.path.exists(self._sink_filename()):
            self._after_load_sink_data_failure(warn=False)
            return

        self._sink_family = family.get_family(config_parser.get('ramses', 'type-sink'))

        with open(self._sink_filename()) as sink_file:
            reader = csv.reader(sink_file, skipinitialspace=True)
            data = list(reader)

        if len(data)<2:
            self._after_load_sink_data_failure()
            return

        column_names = data[0]
        dimensions = data[1]
        data = np.array(data[2:], dtype=object)

        if column_names[0][0]!='#' or dimensions[0][0]!='#':
            self._after_load_sink_data_failure()
            return

        self._fix_fortran_missing_exponent(data)

        column_names[0] = column_names[0][1:].strip()
        dimensions[0] = dimensions[0][1:].strip()

        self._sink_column_names = [self._translate_array_name(c, reverse=True)
                                   for c in column_names]
        self._sink_dimensions = dimensions
        self._sink_data = data

    def _after_load_sink_data_failure(self, warn=True):
        if warn:
            warnings.warn("Unexpected format in file %s -- sink data has not been loaded" % self._sink_filename())

        self._sink_column_names = self._sink_dimensions = self._sink_data = []
        self._sink_family = None


    @staticmethod
    def _fix_fortran_missing_exponent(data_array):
        flattened_data = data_array.flat
        for i in range(len(flattened_data)):
            d = flattened_data[i]
            if "-" in d and "E" not in d:
                flattened_data[i] = "E-".join(d.split("-"))


    def _transfer_sink_data_to_family_array(self):
        if len(self._sink_data)==0:
            return

        target = self[self._sink_family]
        for column, dimension, data in zip(self._sink_column_names,
                                           self._sink_dimensions,
                                           self._sink_data.T):
            dtype = np.float64
            if column=="id":
                dtype = np.int64
            target[column] = data.astype(dtype)
            unit = dimension.replace("m","g").replace("l","cm").replace("t","yr")
            if unit=="1":
                target[column].units="1"
            else:
                target[column].set_units_like(unit)



    def _count_sink_particles(self):
        return len(self._sink_data)



    def _count_particles_using_implicit_families(self):
        """Returns ndm, nstar where ndm is the number of dark matter particles
        and nstar is the number of star particles."""

        npart = 0
        nstar = 0

        dm_i0 = 0
        star_i0 = 0

        self._particle_family_ids_on_disk = []
        self._particle_file_start_indices = []

        if not os.path.exists(self._particle_filename(1)):
            return 0, 0

        if not self._particle_blocks_are_explictly_known:
            distinguisher_field = int(particle_distinguisher[0])
            distinguisher_type = TYPE_MAP[particle_distinguisher[1]]
        else:
            # be more cunning about finding the distinguisher field (likely 'age') -
            # as it may have moved around in some patches of ramses

            distinguisher_name = particle_blocks[int(particle_distinguisher[0])]
            try:
                distinguisher_field = self._particle_blocks.index(distinguisher_name)
            except ValueError:
                # couldn't find the named distinguisher field. Fall back to using index.
                distinguisher_field = int(particle_distinguisher[0])
                if len(self._particle_blocks)>distinguisher_field:
                    pb_name = "%r"%self._particle_blocks[distinguisher_field]
                else:
                    pb_name = "at offset %d"%distinguisher_field
                warnings.warn("Using field %s to distinguish stars. If this is wrong, try editing your config.ini, section [ramses], entry particle-distinguisher."%pb_name)
            distinguisher_type = self._particle_types[distinguisher_field]

        results = remote_map(self.reader_pool,
                             _cpui_count_particles_with_implicit_families,
                             [self._particle_filename(i) for i in self._cpus],
                             [distinguisher_field] * len(self._cpus),
                             [distinguisher_type] * len(self._cpus))

        for npart_this, nstar_this, family_ids in results:
            self._particle_file_start_indices.append((dm_i0, star_i0))
            dm_i0 += (npart_this - nstar_this)
            star_i0 += nstar_this
            npart += npart_this
            nstar += nstar_this

            self._particle_family_ids_on_disk.append(family_ids)

        return npart - nstar, nstar

    def _count_gas_cells(self):
        ncells = remote_map(self.reader_pool, _cpui_count_gas_cells,
                            [self._cpui_level_iterator_args(xcpu) for xcpu in self._cpus])
        self._gas_i0 = np.cumsum([0] + ncells)[:-1]
        return np.sum(ncells)

    def _cpui_level_iterator_args(self, cpu=None):
        if cpu:
            return cpu, self._amr_filename(cpu), self._info['ordering type'] == 'bisection', self._maxlevel, self._ndim
        else:
            return [self._cpui_level_iterator_args(x) for x in self._cpus]

    def _level_iterator(self):
        """Walks the AMR grid levels on disk, yielding a tuplet of coordinates and
        refinement maps and levels working through the available CPUs and levels."""

        for cpu in self._cpus:
            yield from _cpui_level_iterator(*self._cpui_level_iterator_args(cpu))

    def _load_gas_pos(self):
        self.gas['pos'].set_default_units()
        smooth = self.gas['smooth']
        smooth.set_default_units()

        boxlen = self._info['boxlen']

        remote_map(self.reader_pool,
                   _cpui_load_gas_pos,
                   [self.gas['pos']] * len(self._cpus),
                   [self.gas['smooth']] * len(self._cpus),
                   [self._ndim] * len(self._cpus),
                   [boxlen] * len(self._cpus),
                   self._gas_i0,
                   self._cpui_level_iterator_args())

    def _load_gas_vars(self, mode=_gv_load_hydro):
        dims = []

        for i in [self._hydro_blocks, self._grav_blocks, self._rt_blocks][mode]:
            if i not in self.gas:
                self.gas._create_array(i)
            if self._ndim < 3 and i[-1] == 'z':
                continue
            if self._ndim < 2 and i[-1] == 'y':
                continue
            dims.append(self.gas[i])
            self.gas[i].set_default_units()

        if not os.path.exists(self._hydro_filename(1)):
            #Case where force_gas = True, make sure rho is non-zero and such that mass=1.
            # This does not keep track of units for mass or rho since their value is enforced.
            logger.info("No hydro file found, gas likely from force_gas=True => hard setting rho gas")
            self.gas['rho'][:] = 1.0
            self.gas['rho'] /= (np.array(self.gas['smooth']) ** 3)
            self.gas['rho'].set_default_units()


        _grid_info_iter = self._level_iterator()

        logger.info("Loading %s files", ['hydro', 'grav', 'rt'][mode])

        filenamer = [self._hydro_filename, self._grav_filename, self._rt_filename][mode]

        remote_map(self.reader_pool,
                   _cpui_load_gas_vars,
                   [dims] * len(self._cpus),
                   [self._maxlevel] * len(self._cpus),
                   [self._ndim] * len(self._cpus),
                   [filenamer(i) for i in self._cpus],
                   self._cpus,
                   self._cpui_level_iterator_args(),
                   self._gas_i0,
                   [mode] * len(self._cpus))

        if mode is _gv_load_gravity:
            # potential is awkwardly in expected units divided by box size
            self.gas['phi'] *= self._info['boxlen']

        logger.info("Done")

    def _iter_particle_families(self):
        for f in self.families():
            if f is not family.gas and f is not self._sink_family:
                yield f

    def _create_array_for_particles(self, name, type_):
        for f in self._iter_particle_families():
            if name not in list(self[f].keys()):
                for f in self._iter_particle_families():
                    self[f]._create_array(name, dtype=type_)


    def _load_particle_block(self, blockname):
        offset = self._particle_blocks.index(blockname)
        _type = self._particle_types[offset]

        self._create_array_for_particles(blockname, _type)

        arrays = []
        for f in self._iter_particle_families():
            arrays.append(self[f][blockname])


        try:
            remote_map(self.reader_pool,
                       _cpui_load_particle_block,
                       [self._particle_filename(i) for i in self._cpus],
                       [arrays] * len(self._cpus),
                       [offset] * len(self._cpus),
                       self._particle_file_start_indices,
                       [_type] * len(self._cpus),
                       self._particle_family_ids_on_disk
                       )
        except Exception:
            warnings.warn("Exception encountered while reading %r; is there an incompatibility in your Ramses configuration?"%blockname)
            del self[blockname]
            raise

        # The potential is awkwardly not in physical units, but in
        # physical units divided by the box size. This was different
        # in an intermediate version, but then later made consistent
        # with the AMR phi output. So, we need to make a correction here
        # IF we are dealing with the latest ramses format.

        if self._particle_blocks_are_explictly_known and blockname == 'phi':
            for f in self._iter_particle_families():
                self[f]['phi'] *= self._info['boxlen']

    def _load_particle_cpuid(self):
        ind0_dm = 0
        ind0_star = 0
        for i, fam_mask in zip(self._cpus, self._particle_family_ids_on_disk):
            mask_dm = fam_mask == 0
            mask_st = fam_mask == 1

            ind1_dm = ind0_dm + mask_dm.sum()
            ind1_star = ind0_star + mask_st.sum()

            self.dm['cpu'][ind0_dm:ind1_dm] = i
            self.star['cpu'][ind0_star:ind1_star] = i
            ind0_dm, ind0_star = ind1_dm, ind1_star

    def _load_gas_cpuid(self):
        gas_cpu_ar = self.gas['cpu']
        i1 = 0
        for coords, refine, cpu, level in self._level_iterator():
            for cell in range(2 ** self._ndim):
                i0 = i1
                i1 = i0 + (refine[cell] == 0).sum()
                gas_cpu_ar[i0:i1] = cpu

    def loadable_keys(self, fam=None):

        if fam is None:
            keys = None
            for f0 in self.families():
                if keys:
                    keys.intersection_update(self.loadable_keys(f0))
                else:
                    keys = set(self.loadable_keys(f0))
        else:
            if fam is family.gas:
                keys = ['x', 'y', 'z', 'smooth'] + self._hydro_blocks + self._rt_blocks
            elif fam in self._iter_particle_families():
                keys = self._particle_blocks
            elif fam is self._sink_family:
                keys = set(self._sink_column_names)
            else:
                keys = set()

        keys_ND = set()
        for key in keys:
            keys_ND.add(self._array_name_1D_to_ND(key) or key)
        return list(keys_ND)

    @property
    def is_cosmological(self):
        return not self._not_cosmological()

    @property
    def is_not_cosmological(self):
        return self._not_cosmological()

    def _not_cosmological(self):
        not_cosmological = True

        if "cosmo" in self._namelist and self._namelist["cosmo"]:
            not_cosmological = False

        if not self._namelist:
            warnings.warn("Namelist file either not found or unable to read. " +
                          "Guessing whether run is cosmological from cosmological parameters assuming flat LCDM.")
            not_cosmological = (self._info['omega_k'] == self._info['omega_l'] == 0)

        return not_cosmological

    def _convert_tform(self):
        # Copy the existing t array in weird Ramses format into a hidden raw array
        self.star['tform_raw'] = self.star['tform']
        self.star['tform_raw'].units = self._file_units_system[1]

        if self.is_cosmological:
            # Only attempt tform conversion for cosmological runs. The built-in tforms for isolated runs
            # are actually meaningful (issue 554)
            from ..analysis import ramses_util

            # Replace the tform array by its usual meaning using the birth files
            ramses_util.get_tform(self, times_are_proper=self.times_are_proper)


    @property
    def times_are_proper(self) -> bool:
        """Best guess for whether the times in the output (e.g. star formation times) are proper times"""
        if hasattr(self, "_is_using_proper_time"):
            # Already set, skipping
            pass
        elif config_parser.has_option("ramses", "proper_time"):
            self.times_are_proper = config_parser.getboolean(
                "ramses", "proper_time"
            )
        elif self.is_not_cosmological:
            self.times_are_proper = True
        else:
            # If we detect an RT file, assume proper time
            iout = self._timestep_id
            rt_file_candidates = (
                os.path.join(
                    self._dirname,
                    f"rt_{iout}.out{icpu:05d}",
                )
                for icpu in self._cpus
            )
            for rt_file in rt_file_candidates:
                if not os.path.exists(rt_file):
                    continue
                choice = "proper"
                reason = f"one RT file was detected ({rt_file})"
                self.times_are_proper = True
                break
            else:
                choice = "conformal"
                reason = "no RT file was found"
                self.times_are_proper = False
            warnings.warn(
                f"Assumed times to be in {choice} units because {reason}. "
                "If this is incorrect, pass the `times_are_proper` keyword "
                "argument when loading the dataset, or set the option `proper_time` "
                "in your .pynbodyrc."
            )

        return self._is_using_proper_time

    @times_are_proper.setter
    def times_are_proper(self, value):
        self._is_using_proper_time = value

    def _convert_metal_name(self):
        # Name of ramses metallicity has no 's' at the end, contrary to tipsy and gadget
        # Correcting this prevents analysis routines relying on 'metals' field from breaking.
        self.star['metals'] = self.star['metal']

    def _load_array(self, array_name, fam=None):
        # Framework always calls with 3D name. Ramses particle blocks are
        # stored as 1D slices.

        if array_name == 'cpu':
            self['cpu'] = np.zeros(len(self), dtype=int)
            self._load_particle_cpuid()
            self._load_gas_cpuid()

        elif fam is not family.gas and fam is not None:

            # Deal with tform for stars that require extra conversion
            if array_name in ('tform', 'tform_raw'):
                if 'tform' in self._particle_blocks: # Only attempt these conversion if tform is actually on disc (Issue #689)
                    self._load_particle_block('tform')
                    self._convert_tform()

            # Deal with metals for stars that have extra name mapping
            elif array_name in ('metals', 'metal'):
                if 'metal' in self._particle_blocks:
                    self._load_particle_block('metal')
                    self._convert_metal_name()

            elif array_name in self._split_arrays:
                for array_1D in self._array_name_ND_to_1D(array_name):
                    self._load_particle_block(array_1D)

            elif array_name in self._particle_blocks:
                self._load_particle_block(array_name)
            else:
                raise OSError("No such array on disk")

        elif fam is family.gas:

            if array_name == 'pos' or array_name == 'smooth':
                if 'pos' not in self.gas:
                    self.gas._create_array('pos', 3)
                if 'smooth' not in self.gas:
                    self.gas._create_array('smooth')
                self._load_gas_pos()
            elif array_name == 'vel' or array_name in self._hydro_blocks:
                self._load_gas_vars()
            elif array_name in self._grav_blocks:
                self._load_gas_vars(1)
            elif array_name in self._rt_blocks_3d:
                warnings.warn("Loading RT data from disk. Photon densities are stored in flux units by Ramses and need "
                              "to be multiplied by the reduced speed of light of the run to obtain a physical number. "
                              "This is currently left to the user, see issue 542 for more discussion.")
                self._load_gas_vars(_gv_load_rt)
                for u_block in self._rt_blocks_3d:
                    self[fam][u_block].units = self._rt_unit
            elif array_name == 'mass':
                # Very special case. If 'mass' has been created already for particle blocks (or sink blocks), and the
                # user requests it across the whole simulation, the SimSnap framework will issue this request rather
                # than a request to load across all families (handled below). If we raise an exception, the user
                # will see a confusing message about mass being unavailable for gas. So instead we redirect this to
                # a derived array, which will result in a full mass array being returned (good) and a warning about
                # conjoining derived and non-derived arrays (logical, at least).
                self._derive_array('mass', fam)
            else:
                raise OSError("No such array on disk")
        elif fam is None and array_name in ['pos', 'vel']:
            # synchronized loading of pos/vel information
            if 'pos' not in self:
                self._create_array('pos', 3)
            if 'vel' not in self:
                self._create_array('vel', 3)
            if 'smooth' not in self.gas:
                self.gas._create_array('smooth')

            if len(self.gas) > 0:
                self._load_gas_pos()
                self._load_gas_vars()

            # the below triggers loading ALL particles, not just DM
            for name in 'x','y','z','vx','vy','vz':
                self._load_particle_block(name)

        elif fam is None and array_name == 'mass':
            # mass has to be handled separately because it is present as a particle block
            # but not as a gas block -- there it will be derived. It is also present as a sink block,
            # which is again handled differently. So somehow we need to bring all of this together in
            # a way which doesn't baffle users.

            self._create_array('mass')
            self._load_particle_block('mass')
            self['mass'].set_default_units()

            if len(self.gas) > 0:
                gasmass = mass(self.gas)
                gasmass.convert_units(self['mass'].units)
                self.gas['mass'] = gasmass
        else:
            raise OSError("No such array on disk")



    @staticmethod
    def _can_load(f):
        tsid = _timestep_id(f)
        if tsid:
            return os.path.isdir(f) and os.path.exists(os.path.join(f, f"info_{tsid}.txt"))
        return False


@RamsesSnap.decorator
def _translate_info(sim):

    if sim._info['H0']>1e-3:
        sim.properties['a'] = sim._info['aexp']
        sim.properties['omegaM0'] = sim._info['omega_m']
        sim.properties['omegaL0'] = sim._info['omega_l']
        sim.properties['h'] = sim._info['H0'] / 100.

    # N.B. these conversion factors provided by ramses already have the
    # correction from comoving to physical units
    d_unit = sim._info['unit_d'] * units.Unit("g cm^-3")
    t_unit = sim._info['unit_t'] * units.Unit("s")
    l_unit = sim._info['unit_l'] * units.Unit("cm")

    sim.properties['boxsize'] = sim._info['boxlen'] * l_unit

    if sim.is_not_cosmological:
        sim.properties['time'] = sim._info['time'] * t_unit
    else:
        sim.properties["time"] = age(sim, unit="Gyr") * units.Unit("Gyr")

    sim._file_units_system = [d_unit, t_unit, l_unit]


@RamsesSnap.derived_array
def mass(sim):
    """Gas mass estimated from cell size and density"""
    return sim['rho'] * sim['smooth'] ** 3


@RamsesSnap.derived_array
def temp(sim):
    """Gas temperature derived from pressure and density """

    # Has to be redefined and rederived here from Ramses native variables
    # to avoid running into circular dependencies with the traditional derived definition
    # Now uses the self-consistent molecular weight field from pynbody (issue 598)

    from ..derived import mu
    mu_est = array.SimArray(np.ones(len(sim)), units="1")
    for i in range(5):
        temp = ((sim['p'] / sim['rho']) * (mu_est * units.m_p / units.k)).in_units("K")
        mu_est = mu(sim, t0=temp)
    return temp



import warnings

import numpy as np
import numpy.testing as npt
import pytest

import pynbody

# load Alan Duffy's module from https://bitbucket.org/astroduff/pyreadgadget
from pynbody.test_utils.pyread_gadget_hdf5 import pyread_gadget_hdf5


@pytest.fixture
def snap():
    return pynbody.load('testdata/gadget3/data/subhalos_103/subhalo_103')

@pytest.mark.parametrize('load_all', (True, False))
def test_halo_loading(snap, load_all) :
    """ Check that halo loading works """
    h = pynbody.halo.subfindhdf.SubFindHDFHaloCatalogue(snap)

    if load_all:
        h.load_all()


    assert len(h) == 4226
    assert len(h[0]) == 15682
    assert len(h[10]) == 2052
    assert (h[0]['iord'][::100] == [1992767, 1910973, 1977535, 1878091, 1829826, 1796672, 1994317,
          1977404, 1846089, 1993277,   26437,   26444, 2091972, 1961924,
          2026691, 1879228,   43338, 1994823,    9929, 1878477, 1975888,
          2025924, 2073669, 1811915,   59852, 1993403, 1895483, 2010435,
          1779272,   43725, 1796425, 2024514, 2043206, 1960125, 1959244,
          1977162, 1926859, 1943632, 2057932, 2026705,   10062, 2058055,
          1975238, 2023881, 1893316, 4074692, 3991240, 4057412, 4040134,
          4156105, 3976261, 4156366, 4041918, 4008000, 4090186, 4122316,
          3959240, 4139596, 4058693, 4008253, 4188876, 4090959, 4154698,
          4106436, 3975619, 4058439, 4188104, 4107217, 4170954, 3943750,
          4089420, 3926590, 4188108, 3976772, 3925836, 4090449, 4073937,
          3992650, 4140492, 2107470, 4171336, 4058952, 3958859, 4058043,
          2124365, 4089929, 4073282, 3893191, 4107334, 4172356, 3991626,
          4074172, 4140357, 4009019, 4139841, 4042173, 4072779, 4090556,
          4139079, 3958478, 4107455, 4123204, 3860548, 4009276, 4042560,
          3925315, 4057148, 3975119, 4188097, 4123582, 4057149, 4189386,
          4072391, 4009790, 3894209, 4022986, 2140232, 3976894, 4139466,
          4139720, 4172109, 4137416, 3992521, 1895749, 1961287, 1943873,
          2009540, 1895362, 1878718,   10951, 1894084, 1942726, 1878852,
          1861185, 2025542, 2010179, 2090444, 1961408, 1910854, 1861443,
          1878342, 1912000, 1862595, 2059211, 1942987, 1976655, 1994058,
          2042696, 1992905, 2092106, 2074698, 2026952, 1944013,   10185,
          1960656, 1912012, 2090949]).all()
    assert (h[10]['iord'][::100] == [ 782722,  865539,  865923,  784260,  849150,  750470,  898950,
           866689,  882560, 2913154, 2865029, 2880647, 2896644, 2995974,
          2913929, 2881289, 2896517, 2995969,  832898,  865671,  832639]).all()

    assert h[0].properties['NsubPerHalo'] == 6
    assert np.all(h[0].properties['children'] == [0, 1, 2, 3, 4, 5])

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=RuntimeWarning, message="Accessing multiple halos")
        for i, halo in enumerate(h[0:10]):
            halo['mass'].sum()
            for fam in [halo.g, halo.d, halo.s]:
                assert (len(fam['iord']) ==
                        snap._hdf_files[0][snap._family_to_group_map[fam.families()[0]][0]]['Length'][i])

def test_subhalos(snap):
    """Test that subhalos can be accessed from each parent FOF group, as well as directly through a subs object"""
    h = pynbody.halo.subfindhdf.SubFindHDFHaloCatalogue(snap)
    subs = pynbody.halo.subfindhdf.SubFindHDFHaloCatalogue(snap, subhalos=True)
    h.load_all()

    assert (h[0].subhalos[0]['iord'] == subs[0]['iord']).all()
    assert (h[1].subhalos[1]['iord'] == subs[7]['iord']).all()

    with pytest.warns(DeprecationWarning):
        assert (h[1].sub[1]['iord'] == subs[7]['iord']).all()

def test_deprecated_subs_keyword(snap):
    with pytest.warns(DeprecationWarning):
        new_halos = snap.halos(subs=True)
        assert isinstance(new_halos, pynbody.halo.subfindhdf.SubFindHDFHaloCatalogue)
        assert len(new_halos) == 3294

def test_finds_correct_halo(snap):
    h = snap.halos()
    assert isinstance(h, pynbody.halo.subfindhdf.SubFindHDFHaloCatalogue)
    h = snap.halos(subhalos=True)
    assert isinstance(h, pynbody.halo.subfindhdf.SubFindHDFHaloCatalogue)


def test_grp_array(snap):
    h = snap.halos()
    grp = h.get_group_array()
    for i in range(0,100,10):
        assert len(snap['iord'][grp==i]) == len(h[i])
        assert (h[i]['iord'] == snap['iord'][grp==i]).all()

def test_fof_vs_sub_assignment(snap):
    h = snap.halos()
    file_mass_unit = snap.infer_original_units("g")
    npt.assert_allclose(h.get_dummy_halo(0).properties['Mass'].in_units(file_mass_unit),
                        28.604694074339932)
    npt.assert_allclose(h.get_dummy_halo(0).properties['Halo_M_Crit200'].in_units(file_mass_unit),
                        29.796955896599684)
    npt.assert_allclose(h.get_dummy_halo(1).properties['Mass'].in_units(file_mass_unit),
                        8.880245794949587)
    npt.assert_allclose(h.get_dummy_halo(1).properties['Halo_M_Crit200'].in_units(file_mass_unit),
                        8.116568749712314)

@pytest.mark.parametrize('subhalos', (True, False))
@pytest.mark.parametrize('with_units', (True, False))
def test_properties_all_halos(snap, subhalos, with_units):
    h = snap.halos(subhalos=subhalos)
    properties = h.get_properties_all_halos(with_units=with_units)

    filesub = 'testdata/gadget3/data/subhalos_103/subhalo_103'
    dir = 'subfind' if subhalos else 'fof'
    FoF_Mass = pyread_gadget_hdf5(filesub + '.0.hdf5', 10, 'Mass', sub_dir=dir, nopanda=True, silent=True)
    FoF_CoM = pyread_gadget_hdf5(filesub+'.0.hdf5', 10, 'CenterOfMass', sub_dir=dir, nopanda=True, silent=True)

    npt.assert_allclose(FoF_Mass.ravel(), properties['Mass'])
    npt.assert_allclose(FoF_CoM, properties['CenterOfMass'])

    if not subhalos:
        for i in range(5):
            assert (h[i].properties['children'] == properties['children'][i]).all()

    if with_units:
        npt.assert_allclose(properties['Mass'].units.ratio(snap.infer_original_units("g")), 1.0)
        npt.assert_allclose(properties['CenterOfMass'].units.ratio(snap.infer_original_units("cm")), 1.0)
    else:
        assert not hasattr(properties['Mass'], 'units')

@pytest.mark.filterwarnings("ignore:Accessing multiple halos")
@pytest.mark.parametrize('load_all', (True, False))
def test_halo_values(snap, load_all) :
    """ Check that halo values (and sizes) agree with pyread_gadget_hdf5 """

    filesub = 'testdata/gadget3/data/subhalos_103/subhalo_103'



    FoF_Mass = pyread_gadget_hdf5(filesub+'.0.hdf5', 10, 'Mass', sub_dir='fof', nopanda=True, silent=True)
    FoF_MassType = pyread_gadget_hdf5(filesub+'.0.hdf5', 10, 'MassType', sub_dir='fof', nopanda=True, silent=True)
    Sub_Mass = pyread_gadget_hdf5(filesub+'.0.hdf5', 10, 'Mass', sub_dir='subfind', nopanda=True, silent=True)
    Sub_MassType = pyread_gadget_hdf5(filesub+'.0.hdf5', 10, 'MassType', sub_dir='subfind', nopanda=True, silent=True)
    NsubPerHalo = pyread_gadget_hdf5(filesub+'.0.hdf5', 10, 'NsubPerHalo', sub_dir='subfind', nopanda=True, silent=True)
    OffsetHalo = np.roll(NsubPerHalo.cumsum(), 1)
    OffsetHalo[0]=0 ## To start counter

    h = snap.halos()
    if load_all:
        h.load_all()

    FoF_CoM = pyread_gadget_hdf5(filesub+'.0.hdf5', 10, 'CenterOfMass', sub_dir='fof', nopanda=True, silent=True)
    Sub_CoM = pyread_gadget_hdf5(filesub+'.0.hdf5', 10, 'CenterOfMass', sub_dir='subfind', nopanda=True, silent=True)

    # Check the Halo Array values
    for i,halo in enumerate(h[0:10]) :
        assert(np.allclose(halo.properties['CenterOfMass'], FoF_CoM[i], rtol=1e-3))

        for j, s in enumerate(halo.subhalos) :
            assert(np.allclose(s.properties['CenterOfMass'], Sub_CoM[OffsetHalo[i]+j], rtol=1e-3))

    ###
    # Test the Halo particle information
    ###

    # Mass of each component for FOF halos
    for i,halo in enumerate(h[0:10]) :
        assert(np.allclose(halo.g['mass'].sum(), FoF_MassType[i,0], rtol=1e-3))
        assert(np.allclose(halo.dm['mass'].sum(), FoF_MassType[i,1], rtol=1e-3))
        assert(np.allclose(halo.s['mass'].sum(), FoF_MassType[i,4], rtol=1e-3))
        assert(np.allclose(halo['mass'].sum(), FoF_Mass[i], rtol=1e-3))

    # Mass of each component for Subhalos
    for i,halo in enumerate(h[0:10]) :
        for j, s in enumerate(halo.subhalos) :
            assert(np.allclose(s.g['mass'].sum(), Sub_MassType[OffsetHalo[i]+j,0], rtol=1e-3))
            assert(np.allclose(s.dm['mass'].sum(), Sub_MassType[OffsetHalo[i]+j,1], rtol=1e-3))
            assert(np.allclose(s.s['mass'].sum(), Sub_MassType[OffsetHalo[i]+j,4], rtol=1e-3))
            assert(np.allclose(s['mass'].sum(), Sub_Mass[OffsetHalo[i]+j], rtol=1e-3))

    FoF_Temp = pyread_gadget_hdf5(filesub+'.0.hdf5', 0, 'Temperature', sub_dir='fof', nopanda=True, silent=True, physunits=True)[:, 0]
    FoF_Length = pyread_gadget_hdf5(filesub+'.0.hdf5', 0, 'Length', sub_dir='fof', nopanda=True, silent=True, physunits=True).astype(int)[:, 0]
    FoF_Offset = pyread_gadget_hdf5(filesub+'.0.hdf5', 0, 'Offset', sub_dir='fof', nopanda=True, silent=True, physunits=True).astype(int)[:, 0]

    # Test the Particle Temperature and implicitly the particle ordering
    for i,halo in enumerate(h[0:10]) :
        npt.assert_allclose(
            halo.g['temp'],
            FoF_Temp[FoF_Offset[i]:FoF_Offset[i]+FoF_Length[i]],
        )

@pytest.mark.parametrize('load_all', (True, False))
def test_halo_properties_physical_units(snap, load_all):
    h = snap.halos()
    if load_all:
        h.load_all()
    h.physical_units()
    npt.assert_allclose(h[0].properties['CenterOfMass'], [1242.674894, 1571.460534, 2232.622292])
